from .main_loop import MainLoop
from .cancel_action import CancelAction
from .check_ready_action import CheckReadyAction
from .remove_incorrect_action import RemoveIncorrectJobsAction
from .task_status_updater import TasksStatusUpdater
from .tasks_for_planner_sync_with_tracking import TaskForPlannerSyncWithTracking


from tortoise.utils.audio import load_voices
from torchaudio import save
from uuid import uuid4
from tortoise.utils.wav2vec_alignment import Wav2VecAlignment
import torch
from uuid import uuid4


class Generator:
    def __init__(self, tts):
        self.tts = tts
    def _generate(self, text, voice, count):
        voice_sel = [voice]
        voice_samples, conditioning_latents = load_voices(voice_sel)

        gen, dbg_state = self.tts.tts_with_preset(
            text,
            k=count,
            voice_samples=voice_samples,
            conditioning_latents=conditioning_latents,
            preset='fast',
            use_deterministic_seed=None,
            return_deterministic_state=True,
            cvvp_amount=0
        )

        if not isinstance(gen, list):
            gen = [gen]
        return gen


    def simple(self, text, voice, count):
        print(f'Voiceover simple, text `{text}` with voice `{voice}`')
        gen = self._generate(text, voice, count)
        print('Generated. Saving...')
        fnames = []
        for index, g in enumerate(gen):
            fname = f'{uuid4()}.output.{index}.wav'
            save(f'/stash/{fname}', g.squeeze(0).cpu(), 24000)
            fnames.append(fname)
        return fnames

#    This methiod allows you to get the part of audio for each token in the input
#    It's not very precise as sounds gradually transform into each other
#    Also, there is no real use case right now
#    def alignment(self, output_file_name, text, voice, count):
#        print(f'Voiceover aligned, text `{text}` with voice `{voice}`')
#        alignment = Wav2VecAlignment()
#        gen = self._generate(text, voice, count)
#        print('Generated. Saving...')
#        fnames = []
#        for index, g in enumerate(gen):
#            fname = f"{output_file_name}.aligned.{index}.wav.bin"
#            alignments = alignment.align(g.squeeze(1), text, 24000)
#            data = dict(audio = g, alignment = alignments)
#            torch.save(data, f'/stash/{fname}')
#            fnames.append(fname)
#
#       return fnames


import cv2
from frame import Frame
from settings import AnalysisSettings
from cv2 import cvtColor, COLOR_BGR2GRAY, Laplacian, CV_64F
from PIL import Image

def layer_capture(settings: AnalysisSettings):
    cap = cv2.VideoCapture('/resources/input/'+settings.source_file_name)

    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    print(f"Total amount of frames {total_frames}")
    total_frames_digits = len(str(total_frames))

    print("Starting cap")
    cap_count = 0

    while cap.isOpened():
        ret, src_frame = cap.read()

        if not ret:
            break

        pil_image = Image.fromarray(cv2.cvtColor(src_frame, cv2.COLOR_BGR2RGB))
        frame = Frame(src_frame, pil_image)
        frame.index = cap_count
        frame.timestamp_in_ms = cap.get(cv2.CAP_PROP_POS_MSEC)
        frame.filename = str(cap_count).zfill(total_frames_digits) + '.png'
        frame.laplacian = Laplacian(frame.frame, CV_64F).var()
        yield frame

        cap_count += 1
        print(f"{int(100 * cap_count / total_frames)}% done")


    cap.release()


from .interface import ControllersSetup, ControllerInstanceSetup
from .service import ControllerService, ControllerServiceStatus, ControllerServiceStatus
from ..controller import ControllerRegistry, ISingleLoadableModelApi
from yo_fluq import Query

class Setuper:
    def __init__(self, setup: ControllersSetup, service: ControllerService):
        self.setup = setup.controllers
        self.service = service

    def find(self, statuses: ControllerServiceStatus, setup: ControllerInstanceSetup) -> tuple[ControllerServiceStatus.Controller, ControllerServiceStatus.Instance|None]:
        name = ControllerRegistry.to_controller_name(setup.decider)
        status = Query.en(statuses.containers).where(lambda z: z.name == name).single()
        for instance in status.instances:
            if instance.parameter == setup.parameter:
                return status, instance
        return status, None

    def _install(self, statuses: ControllerServiceStatus):
        changed = False
        for s in self.setup:
            status, _ = self.find(statuses, s)
            if not status.installation_status.installed:
                self.service.install(s.decider)
                changed = True
        return changed


    def _run(self, statuses: ControllerServiceStatus):
        changed = False
        for s in self.setup:
            status, instance = self.find(statuses, s)
            if instance is None:
                self.service.run(s.decider, s.parameter)
                changed = True
        return changed

    def _load_model(self, statuses: ControllerServiceStatus):
        changed = False
        for s in self.setup:
            _, instance = self.find(statuses, s)
            if instance is None:
                raise ValueError(f"The required service {s.decider} failed to start")

            if s.loaded_model is None:
                continue

            controller = self.service.get_controller(s.decider)
            api = controller.find_api(instance.instance_id)
            if not isinstance(api, ISingleLoadableModelApi):
                raise ValueError(f"Setup for controller {s.decider} expects model {s.loaded_model}, but the decider is not ISingleLoadableModelApi")
            if api.get_loaded_model_name() != s.loaded_model:
                api.load_model(s.loaded_model)
                changed = True
        return changed


    def make_all(self):
        statuses = self.service.status()
        if self._install(statuses):
            statuses = self.service.status()
        if self._run(statuses):
            statuses = self.service.status()
        self._load_model(statuses)



from .task import BrainBoxTask
from .postprocessors import DownloadingPostprocessor, IPostprocessor, FilePostprocessor
from .combined_task import BrainBoxCombinedTask
from .prerequisites import IPrerequisite, CacheUploadPrerequisite, ResourcePrerequisite, CombinedPrerequisite
from .one_brainbox_task_factory import IOneBrainBoxTaskFactory
from .command import BrainBoxCommand
from .extended_task import BrainBoxExtendedTask


from abc import ABC, abstractmethod
from .run_configuration import RunConfiguration
from .docker_controller import DockerController
from typing import cast

class INotebookableController(ABC):
    @abstractmethod
    def get_notebook_configuration(self) -> RunConfiguration:
        pass

    def run_notebook(self):
        self.run_with_configuration(self.get_notebook_configuration())





from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, File, INotebookableController
)
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .settings import TortoiseTTSSettings
from pathlib import Path


class TortoiseTTSController(DockerWebServiceController[TortoiseTTSSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split("\n")
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084},
            mount_resource_folders=dict(
                voices='/home/app/tortoise-tts/tortoise/voices',
                hf_models='/home/app/.cache/huggingface/hub/',
                tortoise_models='/home/app/.cache/tortoise/models',
                stash='/stash',
            ),
            #dont_rm=True
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return TortoiseTTSSettings()

    def create_api(self):
        from .api import TortoiseTTS
        return TortoiseTTS()

    def post_install(self):
        should_install = False
        for model in [
            self.resource_folder('hf_models')/'models--facebook--wav2vec2-large-960h',
            self.resource_folder('hf_models')/'models--jbetker--tacotron-symbols',
            self.resource_folder('hf_models')/'models--jbetker--wav2vec2-large-robust-ft-libritts-voxpopuli',
            self.resource_folder('tortoise_models') /'models--Manmay--tortoise-tts'
        ]:
            if not model.is_dir():
                should_install = True
                break
        if should_install:
            self.run_auxiliary_configuration(self.get_service_run_configuration(None).as_service_worker('--install'))


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import TortoiseTTS

        TortoiseTTS.export_voice('test_voice', [Path(__file__).parent/'test_voice.wav']).execute(api)
        text = VOICEOVER_TEXT
        files = api.execute(BrainBoxTask.call(TortoiseTTS).dub(text=text, voice='test_voice'))
        yield TestReport.last_call(api).href('voiceover').result_is_array_of_files(File.Kind.Audio).with_comment("Voiceover")
        for i, fname in enumerate(files):
            file = api.open_file(fname)
            check_if_its_sound(file.content, tc)



DOCKERFILE = f'''
FROM python:3.9

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

{SmallImageBuilder.GIT_CLONE_AND_RESET(
    'https://github.com/neonbjb/tortoise-tts.git',
    'tortoise-tts',
    '8a2563ecabe93c4fb626f876dd0c52c966edef2f',
    install=False
)}

COPY . /home/app/tortoise-tts/

WORKDIR /home/app/tortoise-tts

RUN rm setup.py & pip install --user -e .

ENTRYPOINT ["python", "/home/app/tortoise-tts/main.py"]

'''


DEPENDENCIES = '''
--extra-index-url https://download.pytorch.org/whl/cu121
appdirs==1.4.4
asttokens==3.0.0
attrs==24.3.0
audioread==3.0.1
bleach==6.2.0
blinker==1.9.0
blis==0.7.11
catalogue==2.0.10
certifi==2024.12.14
cffi==1.17.1
charset-normalizer==3.4.0
click==8.1.8
cloudpathlib==0.20.0
confection==0.1.5
cymem==2.0.10
debugpy==1.8.11
decorator==5.1.1
deepspeed==0.8.3
einops==0.4.1
entrypoints==0.4
exceptiongroup==1.2.2
executing==2.1.0
fastjsonschema==2.21.1
ffmpeg==1.4
filelock==3.16.1
Flask==3.1.0
fsspec==2024.12.0
hjson==3.1.0
huggingface-hub==0.27.0
idna==3.10
importlib_metadata==8.5.0
inflect==7.4.0
ipykernel==6.9.2
ipython==8.18.1
ipython-genutils==0.2.0
itsdangerous==2.2.0
jedi==0.19.2
Jinja2==3.1.5
joblib==1.4.2
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-client==7.1.2
jupyter_core==5.7.2
langcodes==3.5.0
language_data==1.3.0
librosa==0.9.1
llvmlite==0.43.0
marisa-trie==1.2.1
markdown-it-py==3.0.0
MarkupSafe==3.0.2
matplotlib-inline==0.1.7
mdurl==0.1.2
mistune==3.0.2
more-itertools==10.5.0
mpmath==1.3.0
murmurhash==1.0.11
nbconvert==5.3.1
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.2.1
ninja==1.11.1.3
notebook==5.7.13
numba==0.60.0
numpy==1.26.4
nvidia-cublas-cu12==12.1.3.1
nvidia-cuda-cupti-cu12==12.1.105
nvidia-cuda-nvrtc-cu12==12.1.105
nvidia-cuda-runtime-cu12==12.1.105
nvidia-cudnn-cu12==8.9.2.26
nvidia-cufft-cu12==11.0.2.54
nvidia-curand-cu12==10.3.2.106
nvidia-cusolver-cu12==11.4.5.107
nvidia-cusparse-cu12==12.1.0.106
nvidia-nccl-cu12==2.19.3
nvidia-nvjitlink-cu12==12.6.85
nvidia-nvtx-cu12==12.1.105
packaging==24.2
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
pillow==11.0.0
platformdirs==4.3.6
pooch==1.8.2
preshed==3.0.9
progressbar==2.5
prometheus_client==0.21.1
prompt_toolkit==3.0.48
psutil==6.1.1
ptyprocess==0.7.0
pure_eval==0.2.3
py-cpuinfo==9.0.0
pycparser==2.22
pydantic==1.9.1
Pygments==2.18.0
python-dateutil==2.9.0.post0
PyYAML==6.0.2
pyzmq==26.2.0
referencing==0.35.1
regex==2024.11.6
requests==2.32.3
resampy==0.4.3
rich==13.9.4
rotary-embedding-torch==0.3.6
rpds-py==0.22.3
safetensors==0.4.5
scikit-learn==1.6.0
scipy==1.13.1
Send2Trash==1.8.3
shellingham==1.5.4
six==1.17.0
smart-open==7.1.0
sounddevice==0.5.1
soundfile==0.12.1
spacy==3.7.5
spacy-legacy==3.0.12
spacy-loggers==1.0.5
srsly==2.5.0
stack-data==0.6.3
sympy==1.13.3
terminado==0.13.3
testpath==0.6.0
thinc==8.2.5
threadpoolctl==3.5.0
tokenizers==0.13.3
torch==2.2.2+cu121
torchaudio==2.2.2+cu121
torchvision==0.17.2+cu121
tornado==4.2
tqdm==4.67.1
traitlets==5.14.3
transformers==4.31.0
triton==2.2.0
typeguard==4.4.1
typer==0.15.1
typing_extensions==4.12.2
Unidecode==1.3.8
urllib3==2.3.0
wasabi==1.1.3
wcwidth==0.2.13
weasel==0.4.1
webencodings==0.5.1
Werkzeug==3.1.3
wrapt==1.17.0
zipp==3.21.0
'''


from typing import Any
import os
import shutil
import traceback
from dataclasses import dataclass
from threading import Thread

from ...common import Loc, FileIO
from ...common.marshalling import endpoint
from ..controller import TestReport, DockerWebServiceApi, ControllerRegistry, IController, IModelDownloadingController
from .interface import IControllerService, ControllerServiceStatus, InstallationReport, ControllersSetup

from .logging import Log, LoggingLocalExecutor, LogLogger
from yo_fluq import Query

@dataclass
class RunningInstallation:
    report: InstallationReport
    controller: IController
    thread: Thread|None = None

    def run(self):
        try:
            self.controller.install()
        except:
            error =  traceback.format_exc()
            self.report.error = error

    def exited(self):
        is_alive = self.thread.is_alive()
        return not is_alive


@dataclass
class ControllerServerSettings:
    registry: ControllerRegistry
    port: int = 8091
    api: Any = None

class ControllerService(IControllerService):
    def __init__(self, settings: ControllerServerSettings):
        self.settings = settings
        self.running_installation: RunningInstallation|None = None

    def get_controller(self, decider: str|type):
        name = ControllerRegistry.to_controller_name(decider)
        return self.settings.registry.get_controller(name)

    @endpoint(url='/controllers/install')
    def install(self, decider: str|type, join: bool = True) -> InstallationReport|None:
        if self.running_installation is not None and not self.running_installation.exited:
            raise ValueError("Another installation is in progress")
        self.running_installation = None

        controller = self.get_controller(decider)
        log = Log()
        controller.context._executor = LoggingLocalExecutor(log)
        controller.context._logger = LogLogger(log)

        installation_report = InstallationReport(controller.get_name(), log)
        installation = RunningInstallation(installation_report, controller)
        installation.thread = Thread(target=installation.run)
        installation.thread.start()
        self.running_installation = installation

        if join:
            return self.join_installation()


    @endpoint(url='/controllers/uninstall')
    def uninstall(self, decider: str|type, purge: bool = False):
        self.get_controller(decider).uninstall(purge)


    @endpoint(url='/controllers/run')
    def run(self, decider: str|type, parameter: str | None = None) -> str:
        if parameter == '':
            parameter = None
        return self.get_controller(decider).run(parameter)


    @endpoint(url='/controllers/stop')
    def stop(self, decider: str|type, instance_id: str):
        self.get_controller(decider).stop(instance_id)


    @endpoint(url='/controllers/join-installation')
    def join_installation(self) -> InstallationReport:
        if self.running_installation is None:
            raise ValueError("No installation is in progress")
        if not self.running_installation.exited():
            self.running_installation.thread.join()

        result = self.running_installation
        self.running_installation = None
        if result.report.error is not None:
            raise ValueError(f"Installation threw an exception\n{result.report.error}")
        return result.report




    @endpoint(url='/controllers/installation-report')
    def installation_report(self) -> InstallationReport:
        if self.running_installation is None:
            raise ValueError("No installation is in progress")
        return self.running_installation.report



    @endpoint(url='/controllers/status')
    def status(self):
        installation_statuses = self.settings.registry.get_installation_statuses()
        containers = []
        for name, installation_status in installation_statuses.items():
            if not installation_status.installed or installation_status.dockerless_controller:
                containers.append(ControllerServiceStatus.Controller(name, installation_status))
                continue

            container = self.get_controller(name)
            instances = []
            for instance_id, parameter in container.get_running_instances_id_to_parameter().items():
                address = None
                api = container.find_api(instance_id)
                if isinstance(api, DockerWebServiceApi):
                    address = api.address
                instances.append(ControllerServiceStatus.Instance(
                    instance_id,
                    parameter,
                    address
                ))

            report_path = (self.settings.registry.locator.self_test_path / name)
            has_report =  report_path.is_file()
            has_errors = False
            if has_errors:
                report = FileIO.read_pickle(report_path)
                if report.error is not None:
                    has_errors = True

            containers.append(ControllerServiceStatus.Controller(
                name,
                installation_status,
                has_report,
                has_errors,
                tuple(instances)
            ))

        installation = None
        if self.running_installation is not None and not self.running_installation.exited():
            installation = self.running_installation.report.name

        return ControllerServiceStatus(
            containers,
            installation
        )

    @endpoint(url='/controllers/self_test')
    def self_test(self, decider: str|type) -> TestReport:
        controller = self.get_controller(decider)
        return controller.self_test(locator=self.settings.registry.locator, api=self.settings.api)


    @endpoint(url='/controllers/delete_self_test')
    def delete_self_test(self, decider: str|type):
        controller = self.get_controller(decider)
        path = Loc.self_test_path / controller.get_name()
        if path.is_file():
            os.unlink(path)


    def _inner_path(self, controller, path: str):
        while path.startswith('/'):
            path = path[1:]
        path = controller.context.resource_folder()/path
        path.relative_to(controller.context.resource_folder_root)
        return path

    @endpoint(url='/resources/list')
    def list_resources(self, decider: str|type, path: str) -> list[str]|None:
        controller = self.get_controller(decider)
        root_path = controller.resource_folder()
        path = self._inner_path(controller, path)
        if not path.is_dir():
            return None
        return (Query
                .folder(path,'**/*')
                .where(lambda z: z.is_file())
                .select(lambda z: str(z.relative_to(root_path)).replace('\\','/'))
                .to_list()
                )

    @endpoint(url='/resources/delete')
    def delete_resource(self, decider: str|type, path: str, ignore_errors: bool = False):
        controller = self.get_controller(decider)
        path = self._inner_path(controller, path)
        if path.is_file():
            os.unlink(path)
        elif path.is_dir():
            shutil.rmtree(path)
        else:
            if not ignore_errors:
                raise ValueError(f"Cannot delete {path} of decider {controller.get_name()}: no such file/folder")

    @endpoint(url="/resources/setup", method='POST')
    def setup(self, setup: ControllersSetup):
        from .setuper import Setuper
        setuper = Setuper(setup, self)
        setuper.make_all()

    @endpoint(url='/resources/download_models', method='POST')
    def download_models(self, decider: str|type, models: list):
        controller = self.get_controller(decider)
        if not isinstance(controller, IModelDownloadingController):
            raise ValueError(f"Decider {decider} is not ModelDownloadingController")
        controller.download_models(models)






from typing import *
from frame import Frame

def layer_bufferer(iterable: Iterable[Frame], buffer_size_in_ms, field, maximize: bool = True):
    best_frame = None
    best_value = None
    start = None

    for frame in iterable:
        value = getattr(frame, field)
        if not maximize:
            value = -value
        if start is None or value>best_value:
            best_value = value
            best_frame = frame
            start = frame.timestamp_in_ms

        if frame.timestamp_in_ms - start > buffer_size_in_ms:
            yield best_frame
            start = None
            best_value = None
            best_frame = None









from .file_system import IFileSystem
from ..executor import LocalExecutor
from typing import *
import os
from pathlib import Path
from .machine import Machine
from .command import Command

class SSHFileSystem(IFileSystem):
    def __init__(self,  machine: Machine):
        self.inner_executor = LocalExecutor()
        self.machine = machine

    def _get_prefix(self):
        return [
            'ssh',
            f'{self.machine.username}@{self.machine.ip_address}'
        ]

    def fix(self, s: Iterable[str]):
        return [ss.replace('\\','/') for ss in s]


    def upload_file(self, local_path: Path, remote_path: str):
        self.create_empty_folder(os.path.dirname(remote_path))
        self.inner_executor.execute(self.fix(['scp', str(local_path), f'{self.machine.username}@{self.machine.ip_address}:{remote_path}']))

    def download_file(self, remote_path: str, local_path: Path):
        os.makedirs(local_path.parent, exist_ok=True)
        self.inner_executor.execute(self.fix(['scp', f'{self.machine.username}@{self.machine.ip_address}:{remote_path}', str(local_path)]))


    def create_empty_folder(self, path: str|Path):
        self.inner_executor.execute(self.fix(self._get_prefix() + ['mkdir', '-p', str(path)]))

    def delete_file_or_folder(self, remote_path: str):
        raise NotImplementedError("")

    def is_file(self, remote_path: str):
        result = self.inner_executor.execute(self._get_prefix()+['test', '-f', remote_path, '&&', 'echo', 'yes', '||', 'echo', 'no'], Command.Options(return_output=True))
        return result==b'yes\n'

    def is_dir(self, remote_path: str):
        result = self.inner_executor.execute(self._get_prefix()+['test', '-d', remote_path, '&&', 'echo', 'yes', '||', 'echo', 'no'], Command.Options(return_output=True))
        return result==b'yes\n'





from .container_runner import IContainerRunner
from .image_source import IImageSource
from .image_builder import IImageBuilder
from .executor import IExecutor, LocalExecutor

class Deployment:
    def __init__(self,
                 source: IImageSource,
                 runner: IContainerRunner|None = None,
                 executor: IExecutor|None = None,
                 builder: None | IImageBuilder = None,
                 build_executor: None|IExecutor = None,
                 ):
        self.builder = builder
        self.source = source
        self.runner = runner
        self.executor = executor if executor is not None else LocalExecutor()
        self.build_executor = build_executor if build_executor is not None else self.executor


    def build(self) -> 'Deployment':
        name = self.source.get_image_name()
        self.builder.build_image(name, self.build_executor)
        return self

    def push(self) -> 'Deployment':
        self.source.push(self.build_executor)
        return self

    def pull(self) -> 'Deployment':
        self.source.pull(self.executor)
        return self

    def obtain_image(self) -> 'Deployment':
        if self.builder is not None:
            return self.build()
        else:
            return self.pull()


    def stop(self, ignore_errors: bool = True) -> 'Deployment':
        containers = self.source.get_relevant_containers(self.executor)
        containers+=(self.source.get_container_name(),)
        for container in containers:
            try:
                self.executor.execute(['docker','stop',container])
            except:
                if not ignore_errors:
                    raise
        return self

    def remove(self, ignore_errors: bool = True) -> 'Deployment':
        containers = self.source.get_relevant_containers(self.executor)
        containers += (self.source.get_container_name(),)
        for container in containers:
            try:
                self.executor.execute(['docker', 'rm', container])
            except:
                if not ignore_errors:
                    raise
        return self


    def run(self) -> 'Deployment':
        self.runner.run(self.source.get_image_name(), self.source.get_container_name(), self.executor)
        return self






from .framework.common import IDecider, ISelfManagingDecider, File
from .framework.brainbox import (
    BrainBoxTask, BrainBoxApi, BrainBoxCombinedTask, IPostprocessor, DownloadingPostprocessor, IBrainBoxTask,
    BrainBoxCommand, FilePostprocessor, ResourcePrerequisite, IPrerequisite, CacheUploadPrerequisite,
    CombinedPrerequisite, BrainBoxExtendedTask, BrainBox, BrainBoxServer, BrainBoxServiceSettings
)
from .framework.controllers import IController, ControllerOverDecider, ControllersSetup
from .framework.job_processing import FailedJobArgument, SimplePlanner, AlwaysOnPlanner
from .framework.media_library import MediaLibrary


from .planner import IPlanner, PlannerArguments, OperatorStateForPlanner
from .planner_action import IPlannerAction
from .start_action import StartCommand
from .stop_action import StopCommand
from .assign_action import AssignAction, FailedJobArgument



from ....framework import BrainBox, TestReport, File
from .workflows import *
from unittest import TestCase
from pathlib import Path


class Test:
    def __init__(self, api: BrainBox.Api, tc: TestCase):
        self.api = api
        self.tc = tc

    def test_generation(self):
        prompt = "cute little cat playing with a woolball, solo, no humans, animal, cat"
        negative_prompt = "bad quality, dull colors, monochrome, boy, girl, people, nsfw, nudity"
        model = 'meinamix_meinaV9.safetensors'
        self.api.execute(TextToImage(
            prompt=prompt,
            negative_prompt=negative_prompt,
            batch_size=2,
            model=model,
            seed=42

        ))
        yield (TestReport
               .last_call(self.api)
               .href('basic-generation')
               .result_is_array_of_files(File.Kind.Image)
               )

        self.api.execute(TextToImage(
            prompt=prompt,
            negative_prompt=negative_prompt,
            batch_size=2,
            lora_01='cat_lora.safetensors',
            model=model
        ))

        yield (TestReport
               .last_call(self.api)
               .href('generation-with-lora')
               .result_is_array_of_files(File.Kind.Image)
               .with_comment("Test to image workflow, with LoRA")
               )


    def test_upscale_and_interrogate(self):
        source_image = File.read(Path(__file__).parent / 'image.png')
        input_name = BrainBox.Task.safe_id() + '.png'
        self.api.upload(input_name, source_image)
        self.api.execute(Upscale(input_name))
        yield (TestReport
               .last_call(self.api)
               .href('upscale')
               .result_is_file(File.Kind.Image)
               .with_uploaded_file(input_name, source_image)
               .with_comment("Upscaling workflow")
               )


        self.api.execute(WD14Interrogate(input_name))
        yield (TestReport
               .last_call(self.api)
               .href('interrogation')
               .with_uploaded_file(input_name, source_image)
               .with_comment("Interrogation with WD14")
               )



    def test_all(self):
        yield TestReport.main_section_content("All ComfyUI functionality is accessible via the same endpoint. The workflow is actvated depending on `workflow_type` parameter")
        yield from self.test_generation()
        yield from self.test_upscale_and_interrogate()



from typing import *
from dataclasses import dataclass
from ..app import IBrainBoxTask, BrainBoxApi
from ...job_processing import Job

@dataclass
class BrainBoxCombinedTask(IBrainBoxTask):
    resulting_task: IBrainBoxTask
    intermediate_tasks: Tuple[IBrainBoxTask, ...] = ()

    def create_jobs(self) -> list[Job]:
        jobs = IBrainBoxTask.to_all_jobs(self.intermediate_tasks)
        if self.resulting_task is not None:
            jobs.extend(self.resulting_task.create_jobs())
        return jobs

    def get_resulting_id(self) -> str|None:
        if self.resulting_task is None:
            return None
        else:
            return self.resulting_task.get_resulting_id()

    def postprocess_result(self, result, api):
        return self.resulting_task.postprocess_result(result, api)

    def before_add(self, api: BrainBoxApi):
        for task in self.intermediate_tasks:
            task.before_add(api)
        if self.resulting_task is not None:
            self.resulting_task.before_add(api)




import copy
from typing import *
from ....framework import IDecider, FailedJobArgument, File, MediaLibrary
from datetime import datetime
from .collector_to_media_library_convertor import CollectorToMediaLibraryConvertor
from .task_builder import TaskBuilder


class Collector(IDecider):

    def to_media_library(self, tags: Dict[str, Dict], **kwargs):
        return CollectorToMediaLibraryConvertor(self.current_job_id, self.cache_folder, tags, kwargs).convert()


    def to_array(self, tags: dict[str, dict], **kwargs):
        final_result = []
        for key, value in tags.items():
            item_result = dict(tags=value, id=key)
            if key not in kwargs:
                raise ValueError(f"Id {key} in tags, but not in kwargs")
            job_result = kwargs[key]
            if isinstance(job_result, FailedJobArgument):
                item_result['result'] = None
                item_result['error'] = job_result.job.error
            else:
                item_result['result'] = job_result
                item_result['error'] = None
            final_result.append(item_result)
        return final_result


    def to_dict(self, **kwargs):
        for key, value in kwargs.items():
            if isinstance(value, FailedJobArgument):
                raise ValueError(f"Dependency failed on argument {key}:\n{value.job.error}")
        return kwargs

    def to_result_array(self, tags: dict[str,dict], raise_if_error: bool = False, **kwargs):
        result = self.to_array(tags, **kwargs)
        final_result = []
        for r in result:
            if r['error'] is not None:
                if raise_if_error:
                    raise ValueError(f"Error for task {r['tags']}:\n{r['error']}")
                continue
            final_result.append(r['result'])
        return final_result



    TaskBuilder = TaskBuilder




from datetime import datetime

import numpy as np

from ....job_processing import Job
from sqlalchemy.orm import Session
from sqlalchemy import select, case, func
from yo_fluq import *
from ....common import HTML

def _calculate_julian_date(dt: datetime):
    julian_date = dt.toordinal() + 1721425.5
    return julian_date

def _select():
    query = select(
        Job.id,
        Job.batch,
        case((
            Job.method.isnot(None),
            Job.decider + ':' + Job.method),
            else_=Job.decider
        ).label("decider"),
        Job.received_timestamp,
        Job.accepted_timestamp,
        Job.finished_timestamp,
        (
                (func.julianday(Job.accepted_timestamp) -
                 func.julianday(Job.received_timestamp)
                 ) * 86400
        ).label('in_queue'),
        case(
            (
                Job.finished,
                (
                        func.julianday(Job.finished_timestamp) -
                        func.julianday(Job.accepted_timestamp)
                ) * 86400,
            ),
            else_ = (
                func.julianday(datetime.now()) -
                func.julianday(Job.accepted_timestamp)
            )*86400
        ).label('in_work'),
        case(
            (Job.success, 'success'),
            (Job.finished, 'failure'),
            (Job.accepted, 'in_work'),
            else_='waiting'
        ).label('status'),
        Job.progress,
        Job.error,
    )
    return query



def _query(session: Session, batch: str):
    query = _select()
    if batch is not None:
        query = query.filter(Job.batch == batch)
    query = query.order_by(Job.received_timestamp.desc())
    result = session.execute(query).all()
    return result

def _pretty_sec(seconds):
    if seconds is None:
        return "None"
    if np.isnan(seconds):
        return "None"
    seconds = int(seconds)
    if seconds<60:
        return f'{seconds}s'
    minutes = seconds//60
    seconds %= 60
    if minutes<60:
        return f'{minutes}m {seconds}s'
    hour = minutes//60
    minutes%=60
    return f'{hour}h {minutes}m {seconds}s'

def _to_html_lines(results):
    html = []
    batches = Query.en(results).feed(fluq.partition_by_selector(lambda z: z['batch'])).to_list()
    for batch in batches:
        first_line = True
        for result in batch:
            html.append('<tr>')
            html.append(f"<td>{result['id']}")
            if result['batch']!=result['id']:
                html.append(f'</br><span style="color:gray;">')
                html.append(result['batch'])
                html.append("</span>")
            html.append(f"<td>{result['decider']}</td>")
            html.append(f"<td>{result['received_timestamp']}")
            if result['in_queue'] is not None:
                html.append(f"<br/>Q: {_pretty_sec(result['in_queue'])}")
            if result['in_work'] is not None:
                html.append(f"<br/>W: {_pretty_sec(result['in_work'])}")
            html.append("</td>")
            html.append(f"<td>{result['status']}")
            if result['status'] == 'in_work':
                html.append(f"<br>T: {_pretty_sec((datetime.now()-result['accepted_timestamp']).total_seconds())}")
                if result['progress'] is not None:
                    html.append(f"<br> C: {int(100*result['progress'])}%")
            html.append("</td>")
            html.append("<td>")
            html.append(HTML.button('/jobs/cancel', "Cancel", dict(job_id=result['id'])))
            html.append("<td>")
            html.append("</tr>")



            if result['error'] is not None:
                html.append('<tr><td colspan=5><tt>')
                html.append(result['error'].replace('\n','<br>'))
                html.append('</tt></td></tr>')

    return "\n".join(html)


def create_batch_page(session, batch):
    records = _query(session, batch)
    records = [r._asdict() for r in records]
    inner = _to_html_lines(records)
    inner = "<table border=1>"+inner+"</table>"
    return HTML.page("Brainbox: Jobs", inner)


import requests
from ....framework import DockerWebServiceApi
from .model import OllamaModel
from .settings import OllamaSettings
from .controller import OllamaController

class Ollama(DockerWebServiceApi[OllamaSettings, OllamaController]):
    def __init__(self, address: str|None = None, parameter: str|None = None):
        super().__init__(address, parameter)


    def completions_json(self, prompt: str, **kwargs):
        reply = requests.post(
            f'http://{self.address}/api/generate',
            json=dict(
                model=self.container_parameter,
                prompt=prompt,
                stream=False,
                **kwargs
            )
        )
        if reply.status_code != 200:
            raise ValueError(f'Status code {reply.status_code}, value\n{reply.text}')
        return reply.json()

    def completions(self, prompt: str, **kwargs):
        return self.completions_json(prompt, **kwargs)['response']

    def question_json(self, prompt: str):
        reply = requests.post(
            f'http://{self.address}/api/chat',
            json=dict(
                model=self.container_parameter,
                messages=[
                    dict(role='user', content=prompt),
                ],
                stream=False
            )
        )
        if reply.status_code != 200:
            raise ValueError(f'Status code {reply.status_code}, value\n{reply.text}')
        return reply.json()

    def question(self, prompt: str):
        return self.question_json(prompt)['message']['content']


    Controller = OllamaController
    Settings = OllamaSettings
    Model = OllamaModel


from typing import *
import requests
from urllib.parse import urlencode
from ....framework import IDecider, File, DockerWebServiceApi
from .settings import OpenTTSSettings
from .controller import OpenTTSController

class OpenTTS(DockerWebServiceApi[OpenTTSSettings, OpenTTSController]):
    def __init__(self, address: str = None):
        super().__init__(address)


    def voiceover(self,
                 text: str,
                 voice: str = 'coqui-tts:en_vctk',
                 lang: str = 'en',
                 speakerId: Optional[str] = 'p256'
             ):
        parameters = dict(text=text, voice=voice, lang=lang, speakerId=speakerId)
        query_string = urlencode(parameters)
        reply = requests.get(f'http://{self.address}/api/tts?' + query_string)
        if reply.status_code != 200:
            raise ValueError(f'OpenTTS server returned {reply.status_code}\n{reply.text}')
        result = File(self.current_job_id+'.output.wav', reply.content, File.Kind.Audio)
        return result

    def __call__(self,
                  text: str,
                  voice: str = 'coqui-tts:en_vctk',
                  lang: str = 'en',
                  speakerId: Optional[str] = 'p256'
                  ):
        return self.voiceover(text,voice,lang,speakerId)

    Controller = OpenTTSController
    Settings = OpenTTSSettings


import pickle
import flask
from ...common.marshalling import Server
from ...common import Loc
from .service import BrainBoxServiceSettings, BrainBoxService
from ...controllers import ControllerServer, ControllerServerSettings
from flask import Flask
from . import html_helpers
from sqlalchemy.orm import Session
import io

class BrainBoxServer(Server):
    def __init__(self, settings: BrainBoxServiceSettings):
        self.settings = settings
        from .api import BrainBoxApi
        controller_settings = ControllerServerSettings(settings.registry, settings.port, BrainBoxApi(f"127.0.0.1:{settings.port}"))
        self._controller_server = ControllerServer(controller_settings)
        self.service = BrainBoxService(settings)
        super().__init__(settings.port, self.service)

    def __call__(self):
        self.service.run()
        super().__call__()

    def bind_app(self, app: Flask):
        self._controller_server.bind_endpoints(app)
        self._controller_server.bind_custom_endpoints(app)
        self.bind_endpoints(app)
        self.bind_heartbeat(app)
        app.add_url_rule('/', view_func=self._main_page, methods=['GET'])

        app.add_url_rule('/cache/upload/', view_func=self._upload_cache_file, methods=['POST'])
        app.add_url_rule('/cache/download/<fname>', view_func=self._download_cache_file, methods=['GET'])


        app.add_url_rule('/html/jobs/main_page', view_func=self._main_page, methods=['GET'])
        app.add_url_rule('/html/jobs/batch_page/<batch_id>', view_func=self._batch_page, methods=['GET'])
        app.add_url_rule('/html/jobs/operator_log', view_func=self._operator_log, methods=['GET'])

    def _main_page(self):
        with Session(self.service.engine) as session:
            return html_helpers.create_main_page(session)

    def _batch_page(self, batch_id: str):
        with Session(self.service.engine) as session:
            return html_helpers.create_batch_page(session, batch_id)

    def _operator_log(self):
        return html_helpers.create_operator_log_page(self.service.get_operator_log())


    def _upload_cache_file(self):
        for key, value in flask.request.files.items():
            value.save(self.settings.locator.cache_folder/key)
        return 'OK'

    def _download_cache_file(self, fname):
        with open(self.settings.locator.cache_folder/fname, 'rb') as file:
            return flask.send_file(
                io.BytesIO(file.read()),
                mimetype='application/octet-stream'
            )



from ..core import OperatorMessage, OperatorState
import traceback
from datetime import datetime
from .decider_log import DeciderLogger
from ...common import IDecider, DeciderContext, File, Locator
from .file_postprocessor import file_postprocess
import time


class DeciderOperator:
    def __init__(self, state: OperatorState):
        self.state = state
        self.current_id: str|None = None

    def cycle(self):
        self.state.logger.decider(self.state.key).event("Starting processing cycle")
        while True:
            if self.state.exit_request:
                break
            success = self.next_task()
            if not success:
                time.sleep(0.01)
        self.state.logger.decider(self.state.key).event("Exiting processing cycle")


    def next_task(self):
        if self.state.jobs_queue.empty():
            if self.state.busy:
                self.state.busy = False
                self.state.not_busy_since = datetime.now()
                self.state.logger.decider(self.state.key).event('Not busy')
            return False

        if not self.state.busy:
            self.state.busy = True
            self.state.logger.decider(self.state.key).event('Busy')

        job = self.state.jobs_queue.get()
        self.current_id = job.id
        self.state.logger.decider(self.state.key).event(f'Processing task {self.current_id}')
        self.state.results_queue.put(OperatorMessage(self.current_id, OperatorMessage.Type.accepted))
        self.state.logger.task(self.current_id).event('Accepted')

        method = job.method
        arguments = job.arguments


        try:
            decider = self.state.api
            decider.context._current_job_id = self.current_id
            decider.context._logger = DeciderLogger(self.current_id, self.state.results_queue)

            if method is not None:
                method_instance = getattr(decider, method)
            elif callable(decider):
                method_instance = decider
            else:
                raise ValueError(f'Method is not set for {self.state.key.decider_name}, and the decider is not callable')

            result = method_instance(**arguments)

            result = file_postprocess(result, decider.cache_folder)

            self.state.results_queue.put(OperatorMessage(self.current_id, OperatorMessage.Type.result, result))
            self.state.logger.task(self.current_id).event('Finished with a success')
            return True
        except:
            msg = f"Error when executing job {self.current_id} for decider {self.state.key}\n" + traceback.format_exc()
            self.state.results_queue.put(OperatorMessage(self.current_id, OperatorMessage.Type.error, msg))
            self.state.logger.task(self.current_id).event('Finished with a failure')
            return False
        finally:
            self.current_id = None



from brainbox.framework import OnDemandDockerApi, FileLike, FileIO, File, RunConfiguration
from .controller import PiperTrainingController
from .settings import PiperTrainingSettings
from .container.settings import TrainingSettings
from pathlib import Path
import zipfile
import os
import shutil

class PiperTraining(OnDemandDockerApi[PiperTrainingSettings, PiperTrainingController]):
    def _unzip_dataset(self, name: str, path: Path):
        csv = []
        dataset_folder = self.controller.resource_folder('trainings', name, 'source')
        shutil.rmtree(dataset_folder, ignore_errors=True)
        os.makedirs(dataset_folder/'wav')
        with zipfile.ZipFile(path, 'r') as file:
            for name in file.namelist():
                character, filename = name.split('/')
                if not filename.endswith('.wav'):
                    continue
                data = file.read(name)
                output_path = dataset_folder / 'wav' / filename
                with open(output_path, 'wb') as output_file:
                    output_file.write(data)
                text = file.read(name.replace('.wav', '.txt')).decode('utf-8')
                csv.append(dict(character=character, id=filename, text=text))
        single_speaker = len(set(c['character'] for c in csv)) == 1
        with open(dataset_folder / 'metadata.csv', 'w') as file:
            for record in csv:
                if single_speaker:
                    file.write(f'{record["id"]}|{record["text"]}\n')
                else:
                    file.write(f'{record["id"]}|{record["character"]}|{record["text"]}\n')


    def execute(self, dataset: FileLike.Type, settings: TrainingSettings|dict|None = None, name: str|None = None):
        if name is None:
            name = FileLike.get_name(dataset).split('.')[0]
        if settings is None:
            settings = {}
        elif isinstance(settings, TrainingSettings):
            settings = settings.__dict__

        path = FileLike(dataset, self.cache_folder).get_path()
        self._unzip_dataset(name, path)
        FileIO.write_json(settings, self.controller.resource_folder('trainings', name)/'settings.json')

        configuration = RunConfiguration(
            detach_and_interactive=False,
            mount_top_resource_folder=True,
            command_line_arguments=['--dataset',name]
        )
        self.controller.run_with_configuration(configuration, print)
        result = []
        result_folder = self.controller.resource_folder('trainings',name,'result')
        for file in os.listdir(result_folder):
            dst_filename = self.current_job_id+'_'+file
            shutil.copy(
                result_folder/file,
                self.cache_folder/dst_filename,
            )
            if file.endswith('.onnx'):
                result.append(dst_filename)
        return result

    def convert_model(self, ckpt_model: FileLike.Type):
        name = FileLike.get_name(ckpt_model)
        print(name)

        with FileLike(ckpt_model, self.cache_folder) as stream:
            with open(self.controller.resource_folder('conversions')/name, 'wb') as output:
                output.write(stream.read())

        configuration = RunConfiguration(
            detach_and_interactive=False,
            mount_top_resource_folder=True,
            command_line_arguments=['--convert',str(name)]
        )
        self.controller.run_with_configuration(configuration, print)




    Controller = PiperTrainingController
    Settings = PiperTrainingSettings
    TrainingSettings = TrainingSettings



from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def resources(test_case: TestCase, api: BrainBox.Api):
    """
    ## Advanced techniques

    ### Browse the resources

    Many deciders have _resources_: files, typically model files, that are passed to the containers and used for inference.
    It is impractical to store them inside containers, and so BrainBox stores them in the host's file system:
    each decider has its own resource folder, which is mounted to the container.

    Aside from models, the training data may be stored in resources instead of file cache, as training data
    are easier to organize with folder structure and predictable filenames.

    This endpoint demonstrates that the server indeed has an access to the resources:
    HelloBrainBox reads them and returns the dictionary of keys set to filenames,
    and values set to file content.
    These resources are created by HelloBrainBox controller as a part of installation procedure.
    """
    resources = api.execute(BrainBox.Task.call(HelloBrainBox).resources())
    test_case.assertDictEqual(
        {'nested/resource': 'HelloBrainBox nested resource', 'resource': 'HelloBrainBox resource'},
        resources
    )

    """
    However, a uniform access to all the resources is provided by API, e.g. to list all the resources:
    """

    resources = api.controller_api.list_resources(HelloBrainBox, '/')
    test_case.assertListEqual(
        ['resource', 'nested/resource'],
        resources
    )


def model_installation(test_case: TestCase, api: BrainBox.Api):
    """
    ### Download a custom model

    Many deciders can run with different models, and these models need to be downloaded.
    All the deciders load some models at the installation time to allow self-test to run.
    However, more models are usually needed.

    Sometimes the containers are able to download these models themselves,
    but often enough the models need to be downloaded separately.

    This action can be triggered via `api`:
    """

    api.controller_api.download_models(HelloBrainBox, [HelloBrainBox.Model('google', 'http://www.google.com')])
    resources = api.controller_api.list_resources(HelloBrainBox, '/models')
    test_case.assertListEqual(['models/google'], resources)










import os

import flask
from uuid import uuid4
import traceback
import whisper
import json
from io import BytesIO

class WhisperApp:
    def __init__(self):
        self.model = None
        self.model_name = None


    def create_app(self):
        app = flask.Flask("Whisper STT Web App")
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/load_model', view_func=self.load_model, methods=['POST'])
        app.add_url_rule('/get_loaded_model', view_func=self.get_loaded_model, methods=['GET'])
        app.add_url_rule('/transcribe', view_func=self.transcribe, methods=['POST'])

        return app

    def index(self):
        return f"{type(self).__name__} is running"

    def _load_model_internal(self, model_name):
        self.model = whisper.load_model(model_name, download_root='/resources/')
        self.model_name = model_name
        return 'OK'

    def load_model(self):
        try:
            model_name = flask.request.json['model']
            return self._load_model_internal(model_name)
        except:
            return traceback.format_exc(), 500

    def get_loaded_model(self):
        return flask.jsonify(self.model_name)


    def transcribe(self):
        fname = f'/resources/input_{uuid4()}.wav'
        try:
            file = flask.request.files['file']
            file.save(fname)

            param_file = flask.request.files['parameters']
            param_str = param_file.stream.read()
            params = json.loads(param_str)

            model = params['model']
            if model is None:
                if self.model_name is None:
                    raise ValueError("No model is provided with a request and no model is pre-loaded on the server")
            else:
                if self.model_name is None or self.model_name != model:
                    self._load_model_internal(model)
            result = self.model.transcribe(fname, **params['parameters'])
            return flask.jsonify(result)
        except:
            return traceback.format_exc(), 500
        finally:
            os.unlink(fname)




from .marshalling_metadata import MarshallingMetadata
from flask import Flask
from .server_binding import ServerBinding


class Server:
    def __init__(self, port: int, *objects):
        self.objects = objects
        self.port = port
        metadata = []
        for object in objects:
            metadata.extend(MarshallingMetadata.get_endpoints_from_object(object))
        self.metadata = tuple(metadata)

    def create_alternative_binding(self, name: str, address: str):
        meta = [m for m in self.metadata if m.get_endpoint_address() == address]
        if len(meta) != 1:
            raise ValueError(f"Too much/none ({len(meta)} endpoints for address {address}")
        return ServerBinding(meta[0], name)

    def bind_endpoints(self, app: Flask):
        for metadata in self.metadata:
            app.add_url_rule(
                metadata.get_endpoint_address(),
                view_func=ServerBinding(metadata),
                methods=[metadata.endpoint.method]
            )

    def bind_heartbeat(self, app: Flask):
        app.add_url_rule('/heartbeat', view_func=self.heartbeat, methods=['GET'])

    def bind_app(self, app: Flask):
        self.bind_endpoints(app)
        self.bind_heartbeat(app)


    def heartbeat(self):
        return 'OK'

    def __call__(self):
        app = Flask('RPC_'+'_'.join(type(o).__name__ for o in self.objects))
        self.bind_app(app)
        app.run('0.0.0.0', self.port)


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask, File,
    OnDemandDockerController
)
from .settings import BoilerplateOnDemandSettings
from pathlib import Path



class BoilerplateOnDemandController(OnDemandDockerController[BoilerplateOnDemandSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_default_settings(self):
        return BoilerplateOnDemandSettings()

    def create_api(self):
        from .api import BoilerplateOnDemand
        return BoilerplateOnDemand()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import BoilerplateOnDemand

        result = api.execute(BrainBoxTask.call(BoilerplateOnDemand).execute('test input'))
        tc.assertEquals('output test input', result)
        yield TestReport.last_call(api).href('run')


DOCKERFILE = f"""
FROM python:3.11

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app

ENTRYPOINT ["python3", "/home/app/main.py"]
"""

DEPENDENCIES = """
python-dotenv==1.0.1
"""


from pathlib import Path
from ...common import File

def _dump(file, cache_folder: Path):
    if isinstance(file, File):
        with open(cache_folder / file.name, 'wb') as stream:
            stream.write(file.content)
            file.content = None
        return file.name
    else:
        return file


def file_postprocess(obj, cache_folder: Path):
    if isinstance(obj, File):
        return _dump(obj, cache_folder)
    elif isinstance(obj, tuple):
        return tuple(_dump(f, cache_folder) for f in obj)
    elif isinstance(obj, list):
        return [_dump(f, cache_folder) for f in obj]
    elif isinstance(obj, dict):
        return {key: _dump(value, cache_folder) for key, value in obj.items()}
    else:
        return obj


from yo_fluq import *
from pathlib import Path
from brainbox import BrainBox
from brainbox.framework import Loc
from brainbox.deciders import Ffmpeg




if __name__ == '__main__':
    path = Path(__file__).parent.parent.parent/'piper_training/files'
    files = Query.folder(path,'*.wav').to_list()
    with BrainBox.Api.Test() as api:
        result = api.execute(BrainBox.Task.call(Ffmpeg).concat(files, '.mp3'))
        api.download(result, Path(__file__).parent/'lina.mp3')



import traceback
import flask
import os
import wave
import json
from vosk import Model, KaldiRecognizer, SetLogLevel
from uuid import uuid4

class VoskApp:
    def __init__(self):
        self.model = None
        self.model_name = None

    def create_app(self):
        app = flask.Flask(type(self).__name__)
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/transcribe', view_func=self.transcribe, methods=['POST'])
        app.add_url_rule('/transcribe/<model_name>', view_func=self.transcribe, methods = ['POST'])
        app.add_url_rule('/load_model', view_func=self.load_model, methods=['POST'])
        app.add_url_rule('/get_loaded_model', view_func=self.get_loaded_model, methods=['GET'])
        return app

    def index(self):
        return "OK"


    def _load_model_internal(self, model_name):
        self.model = Model('/models/'+model_name)
        self.model_name = model_name
        return 'OK'

    def load_model(self):
        try:
            model_name = flask.request.json['model']
            return self._load_model_internal(model_name)
        except:
            return traceback.format_exc(), 500

    def get_loaded_model(self):
        return flask.jsonify(self.model_name)


    def transcribe(self, model_name = None):
        fname = f'/resources/input_{uuid4()}.wav'
        try:
            file = flask.request.files['file']
            file.save(fname)

            if model_name is None:
                if self.model_name is None:
                    raise ValueError("No model is provided with a request and no model is pre-loaded on the server")
            else:
                if self.model_name is None or self.model_name != model_name:
                    self._load_model_internal(model_name)
            wf = wave.open(fname, "rb")
            if wf.getnchannels() != 1 or wf.getsampwidth() != 2 or wf.getcomptype() != "NONE":
                raise ValueError("Audio file must be WAV format mono PCM.")
            rec = KaldiRecognizer(self.model, wf.getframerate())
            rec.SetWords(True)
            results = []

            while True:
                data = wf.readframes(4000)
                if len(data) == 0:
                    break
                if rec.AcceptWaveform(data):
                    part_result = json.loads(rec.Result())
                    results.append(part_result)

            part_result = json.loads(rec.FinalResult())
            results.append(part_result)

            return flask.jsonify(results)
        except:
            return traceback.format_exc(), 500
        finally:
            os.unlink(fname)


import threading
import subprocess
import sys
import pickle
import atexit
import time
from pathlib import Path
from uuid import uuid4
from typing import Callable
from .loc import Loc

class Fork:
    def __init__(self, method: Callable, raise_if_exited: bool = True):
        self.method = method
        self.process: subprocess.Popen|None = None
        self.monitor_thread = None
        self.exception_raised = False
        self.raise_if_exited = raise_if_exited

    def start(self):
        try:
            path = Loc.temp_folder / ('fork_' + str(uuid4()))
            with open(path, 'wb') as stream:
                pickle.dump(self.method, stream)
        except Exception as ex:
            raise ValueError(f"Cannot pickle {self.method} to run in fork") from ex
        self.process = subprocess.Popen(
            [
                sys.executable,
                '-m',
                'brainbox.framework.common.fork_worker',
                str(self.method),
                str(path)
            ],
        )
        atexit.register(self.terminate)

        if self.raise_if_exited:
            self.monitor_thread = threading.Thread(target=self._monitor_process, daemon=True)
            self.monitor_thread.start()

        return self

    def _monitor_process(self):
        while True:
            if self.process.poll() is not None:
                if not self.exception_raised:
                    raise RuntimeError(f"Subprocess for {str(self.method)} exited unexpectedly with code {self.process.returncode}")
                break
            time.sleep(0.1)

    def terminate(self):
        if self.process and self.process.poll() is None:
            self.exception_raised = True
            self.process.terminate()
            self.process.wait()


from abc import ABC, abstractmethod
from .....framework import FileLike, File

class IWorkflow:
    @abstractmethod
    def create_workflow_json(self, job_id: str):
        pass


    @abstractmethod
    def get_input_files(self) -> list[FileLike.Type]:
        pass

    def postprocess(self, files: list[File]):
        return files


import time
from typing import *
from .marshalling_metadata import MarshallingMetadata
from .api_binding import ApiBinding
from ..api_utils import ApiUtils
import requests
import re

def bind_to_api(api_type: Type):
    def decorator(cls):
        metadata = MarshallingMetadata.get_endpoints_from_type(api_type)
        abs_methods = set(cls.__abstractmethods__)
        for meta in metadata:
            if meta.name not in abs_methods:
                raise ValueError(f"{meta.name} is present in API but absent/not marked as abstract in {api_type}")
            abs_methods.remove(meta.name)
        cls.__abstractmethods__ = frozenset(abs_methods)
        cls.__api_endpoints__ = metadata
        return cls
    return decorator


class Api:
    def __init__(self, address: str):
        ApiUtils.check_address(address)
        self.address = address

        for meta in type(self).__api_endpoints__:
            setattr(self, meta.name, ApiBinding(address, meta))


    def wait(self, max_time_in_seconds=10):
        ApiUtils.wait_for_reply(
            f'http://{self.address}/heartbeat',
            max_time_in_seconds,
        )











from typing import *
from dataclasses import dataclass,field
from pathlib import Path
import shutil
import os
import zipfile
from datetime import datetime
import pickle
from ..common import FileIO
from copy import deepcopy
from ..common import File

import pandas as pd
from yo_fluq import Queryable
from copy import copy

DESCRIPTION_FILE_NAME = 'description.pkl'
ERRORS_FILE_NAME = 'errors.pkl'


class MediaLibrary:
    @dataclass(frozen=True)
    class Record:
        filename: str
        holder_location: Path|None
        timestamp: None|datetime = None
        job_id: None|str = None
        tags: Dict[str, Any] = field(default_factory=dict)
        inline_content: Any = None

        def get_full_path(self):
            if self.packed:
                raise ValueError('This is a packed file, no full path exists')
            return self.holder_location/self.filename

        def change_folder(self, host_folder):
            return MediaLibrary.Record(self.filename, host_folder, self.timestamp, self.job_id, self.tags)

        @property
        def unpacked(self):
            return self.inline_content is None and self.holder_location.is_dir()

        @property
        def packed(self):
            return self.inline_content is None and self.holder_location.is_file()

        @property
        def inline(self):
            return self.inline_content is not None

        def get_content(self, zip_file = None):
            if self.inline:
                return self.inline_content
            elif self.unpacked:
                with open(self.holder_location / self.filename, 'rb') as stream:
                    return stream.read()
            else:
                if zip_file is not None:
                    return zip_file.read(self.filename)
                with zipfile.ZipFile(self.holder_location, 'r', zipfile.ZIP_DEFLATED) as zp:
                    return zp.read(self.filename)

        def get_file(self, zip_file=None):
            return File(self.filename, self.get_content(zip_file))


    def __init__(self,
                 records: Tuple['MediaLibrary.Record', ...] = (),
                 errors: Tuple[str,...] = (),
                 info: Any = None
                 ):
        self._records = tuple(records)
        self._errors = tuple(errors)
        self._mapping = {r.filename: r for r in self.records}
        self._info = info

    @property
    def records(self) -> Tuple[Record,...]:
        return self._records

    @property
    def errors(self) -> Tuple[str,...]:
        return self._errors

    @property
    def mapping(self) -> Mapping[str, Record]:
        return self._mapping

    @property
    def info(self) -> Any:
        return self._info

    def clone(self) -> 'MediaLibrary':
        return deepcopy(self)

    def limit_to(self, ids: Iterable[str]):
        ids = set(ids)
        return MediaLibrary(
            tuple(deepcopy(z) for z in self.records if z.filename in ids),
            (),
            self.info
        )


    def __getitem__(self, item: str) -> 'MediaLibrary.Record':
        return self.mapping[item]

    def __iter__(self):
        yield from self.records

    def enumerate_content(self) -> Iterable[Tuple['MediaLibrary.Record',bytes]]:
        non_packed = []
        packed = {}
        for record in self.records:
            if record.packed:
                if record.holder_location not in packed:
                    packed[record.holder_location] = []
                packed[record.holder_location].append(record)
            else:
                non_packed.append(record)
        for record in non_packed:
            yield (record, record.get_content())
        for zp_location, records in packed.items():
            with zipfile.ZipFile(zp_location, 'r', zipfile.ZIP_DEFLATED) as zp:
                for record in records:
                    yield (record, record.get_content(zp))

    def to_df(self):
        df = pd.DataFrame([r.tags for r in self.records])
        for field in ['filename','timestamp','job_id']:
            df[field] = [getattr(r, field) for r in self.records]
        return df


    def save(self, location: Path, with_progress_bar: bool = False):
        with zipfile.ZipFile(location, 'w', zipfile.ZIP_DEFLATED) as zp:
            source = Queryable(self.enumerate_content(), len(self.records))

            records = {}
            for record, content in source:
                if not record.inline:
                    zp.writestr(record.filename, content)
                dct = copy(record.__dict__)
                del dct['holder_location']
                records[record.filename] = dct

            zp.writestr(DESCRIPTION_FILE_NAME, pickle.dumps(records))
            zp.writestr(ERRORS_FILE_NAME, pickle.dumps(list(self.errors)))



    def unzip(self, folder: Path, with_progress_bar: bool = False):
        if os.path.isfile(folder):
            raise ValueError(f'{folder} is a file')
        elif os.path.isdir(folder):
            shutil.rmtree(folder)
        os.makedirs(folder)

        source = Queryable(self.enumerate_content(), len(self.records))

        records = []
        for record, content in source:
            if not record.inline:
                with open(folder/record.filename, 'wb') as stream:
                    stream.write(content)
            records.append(record.change_folder(folder))

        return MediaLibrary(tuple(records), self.errors)




    def merge(self, another: 'MediaLibrary') -> 'MediaLibrary':
        return MediaLibrary(
            self.records + another.records,
            self.errors + another.errors
        )


    @staticmethod
    def _load_zip(zip_file: Path, host_folder: Optional[Path]) -> 'MediaLibrary':
        with zipfile.ZipFile(zip_file, 'r', zipfile.ZIP_DEFLATED) as zp:
            errors = pickle.loads(zp.read(ERRORS_FILE_NAME))
            result = []
            desc = pickle.loads(zp.read(DESCRIPTION_FILE_NAME))
            name_list = zp.namelist()
            for key, value in desc.items():
                if 'inline_content' not in value:
                    value['inline_content'] = None #compatibility assignment
                if value['inline_content'] is None and key not in name_list:
                    raise ValueError(f'File {key} is listed in {DESCRIPTION_FILE_NAME}, but is absent from archive file {zip_file}')

                if value['inline_content'] is not None:
                    location = None
                elif host_folder is None:
                    location = zip_file
                else:
                    location = host_folder

                item = MediaLibrary.Record(holder_location=location, **value)
                result.append(item)
        return MediaLibrary(tuple(result), tuple(errors))

    @staticmethod
    def _load_zips(zip_files: Iterable[Path], host_folder: Optional[Path]) -> 'MediaLibrary':
        result = None
        for zip_file in zip_files:
            addition = MediaLibrary._load_zip(zip_file, host_folder)
            if result is None:
                result = addition
            else:
                result = result.merge(addition)
        return result

    @staticmethod
    def read(*zip_files: Path) -> 'MediaLibrary':
        zip_files = [Path(p) for p in zip_files]
        return MediaLibrary._load_zips(zip_files, None)



    @staticmethod
    def generate(path: Path, tags_array: Dict[str,Dict[str,Any]]):
        folder = path.parent/("tmp_"+path.name)
        os.makedirs(folder, exist_ok=True)
        records = []
        for fname, tags in tags_array.items():
            record = MediaLibrary.Record(
                filename = fname,
                holder_location=folder,
                timestamp=datetime.now(),
                job_id='',
                tags = tags,
                inline_content=None
            )
            records.append(record)
            FileIO.write_json(tags, folder/fname)
        library = MediaLibrary(tuple(records))
        library.save(path)
        shutil.rmtree(folder)







from pathlib import Path
import os

class ResourceFolder:
    def __init__(self, root: Path):
        self._root = root

    def __call__(self, *subfolders):
        folder = self._root
        for subfolder in subfolders:
            if subfolder is not None:
                folder /= subfolder
        os.makedirs(folder, exist_ok=True)
        return folder



from pathlib import Path
from typing import cast
from ....framework import DownloadableModel, IController, RunConfiguration
from dataclasses import dataclass
import shutil
import requests

@dataclass
class RhasspyKaldiModel(DownloadableModel):
    language: str

    def get_local_location(self, controller: IController) -> Path:
        profile_folder = controller.resource_folder('profiles')
        folder = profile_folder / self.language
        return folder

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import RhasspyKaldiController
        controller = cast(RhasspyKaldiController, controller)
        for model in models:
            model = cast(RhasspyKaldiModel, model)
            controller.context.logger.log(f"Downloading models for language {model.language}: running endpoint")
            shutil.copy(
                Path(__file__).parent / 'files/default_profile.json',
                controller.resource_folder('profiles', model.language) / 'profile.json'
            )
            cfg = RunConfiguration(
                None,
                mount_resource_folders={'profiles': '/profiles', 'models': '/models'},
                publish_ports={controller.settings.connection.port: 12101},
                custom_flags=['-v', '/etc/localtime:/etc/localtime:ro'],
                command_line_arguments=['--rhasspy', '--language', model.language],
            )
            instance_id = controller.run_with_configuration(cfg)
            controller.wait_for_boot()
            controller.context.logger.log("Endpoint booted, requesting download")
            requests.post(f'http://127.0.0.1:{controller.settings.connection.port}/api/download-profile')
            controller.stop(instance_id)
            path = controller.resource_folder('profiles')/model.language
            if not path.is_dir():
                raise ValueError(f"Error when trying to download the model for language {model.language}"
                                 "The endpoint existed successfully, but the folder is not where it's expected")
            controller.context.logger.log("All good")






from algorithm import VideoProcessorApp
import argparse
from layer_semantic_comparator import get_comparator

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--install', '-i', action='store_true')
    args = parser.parse_args()

    if args.install:
        get_comparator()
        exit(0)


    app = VideoProcessorApp()
    app.run()




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, SmallImageBuilder,
    IImageBuilder, DockerWebServiceController, BrainBoxApi, IModelDownloadingController, DownloadableModel,
    TestReport, INotebookableController
)
from .settings import RhasspyKaldiSettings
from .model import RhasspyKaldiModel
from pathlib import Path


class RhasspyKaldiController(
    DockerWebServiceController[RhasspyKaldiSettings],
    IModelDownloadingController,
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            None
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return RhasspyKaldiModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            None,
            mount_resource_folders={'profiles': '/profiles', 'models': '/models'},
            publish_ports={self.settings.connection.port: 8084}
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return RhasspyKaldiSettings()

    def create_api(self):
        from .api import RhasspyKaldi
        return RhasspyKaldi()

    def post_install(self):
        self.download_models(self.settings.languages)
        
    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .tests import english, german, english_custom
        yield TestReport.attach_source_file(english)
        yield from english(api, tc)
        yield from german(api, tc)
        yield from english_custom(api, tc)


DOCKERFILE = f'''
FROM rhasspy/rhasspy

RUN /usr/lib/rhasspy/.venv/bin/pip install notebook flask

COPY . /home/app

ENTRYPOINT ["/usr/lib/rhasspy/.venv/bin/python", "/home/app/main.py"] 
'''





import subprocess
import argparse
import json
import sys
import time
from pathlib import Path

import toml

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-g','--gradio', action='store_true')
    parser.add_argument('-t','--train')
    parser.add_argument('-m','--mock', action='store_true')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.gradio:
        subprocess.call(["python3", "kohya_gui.py", "--listen", "0.0.0.0", "--server_port", "7860", "--headless"])
        exit(0)

    if args.train is not None:
        path = Path('/app/training/'+args.train)
        with open(path/'parameters.json') as file:
            params = json.load(file)
        if not args.mock:
            error_code = subprocess.call(params)
            with open(path/'error_code.txt', 'w') as error_code_file:
                error_code_file.write(str(error_code))
        else:
            with open(path/'model/config.toml') as cfg_file:
                cfg = toml.load(cfg_file)
            output_path = Path(cfg['output_dir'])
            for interfix in ['', '-000000']:
                output_file=output_path/(cfg['output_name']+interfix+'.safetensors')
                with open(output_file,'wb') as file:
                    file.write(b'koyha_ss_mock_output_file')


            for i in range(100):
                print(f'steps: {i}%', file=sys.stderr)
                time.sleep(0.01)

            with open(path/'error_code.txt', 'w') as error_code_file:
                error_code_file.write("0")















from typing import *
from abc import ABC, abstractmethod
from .command import Command
from pathlib import Path
from .file_system import IFileSystem
from .machine import Machine

class IExecutor(ABC):
    @abstractmethod
    def execute_command(self, command: Command):
        pass

    @abstractmethod
    def get_fs(self) -> IFileSystem:
        pass

    @abstractmethod
    def get_machine(self) -> Machine:
        return Machine.local()


    def execute(self,
                command: Command|Iterable[Union[str,Path]],
                options: Command.Options|None = None
                ):
        if isinstance(command, Command):
            return self.execute_command(command)
        command = Command(tuple(command), options)
        return self.execute_command(command)






from typing import *
from ...common import Logger, Loc
from .resource_folder import ResourceFolder
from ...deployment import Machine, LocalExecutor, IExecutor
from pathlib import Path


TSettings = TypeVar("TSettings")


class ControllerContext(Generic[TSettings]):
    def __init__(self, name: str, settings: TSettings):
        self._name = name
        self.settings = settings
        self._resource_folder_root : Path | None = None
        self._logger: Logger | None = None
        self._machine: Machine|None = None
        self._executor: IExecutor | None = None

    @property
    def resource_folder_root(self) -> Path:
        if self._resource_folder_root is None:
            return Loc.resources_folder
        return self._resource_folder_root

    @property
    def resource_folder(self) -> ResourceFolder:
        return ResourceFolder(self.resource_folder_root/self._name)

    @property
    def logger(self):
        if self._logger is None:
            return Logger()
        return self._logger


    @property
    def machine(self) -> Machine:
        if self._machine is None:
            return Machine.local()
        return self._machine

    @property
    def executor(self) -> IExecutor:
        if self._executor is None:
            return LocalExecutor()
        return self._executor





import torch
import torchaudio
from zonos.model import Zonos
from zonos.conditioning import make_cond_dict

class ZonosModel:
    def __init__(self):
        self.model = Zonos.from_pretrained("Zyphra/Zonos-v0.1-transformer", device="cuda")

    def train(self, path_to_file):
        wav, sampling_rate = torchaudio.load(path_to_file)
        speaker = self.model.make_speaker_embedding(wav, sampling_rate)
        return speaker

    def voiceover(self, text, speaker_model, language, output_file):
        cond_dict = make_cond_dict(
            text=text,
            speaker=speaker_model,
            language=language
        )
        conditioning = self.model.prepare_conditioning(cond_dict)
        codes = self.model.generate(conditioning)
        wavs = self.model.autoencoder.decode(codes).cpu()
        torchaudio.save(output_file, wavs[0], self.model.autoencoder.sampling_rate)



import traceback
import flask
import os

class HelloBrainBoxApp:
    def __init__(self, setting: str, model: str):
        self.setting = setting
        self.model = model


    def create_app(self):
        app = flask.Flask(type(self).__name__)
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/decide', view_func=self.decide, methods = ['POST'])
        app.add_url_rule('/model', view_func=self.get_model, methods=['GET'])
        app.add_url_rule('/resources', view_func=self.resources, methods=['GET'])
        return app

    def index(self):
        return "OK"

    def get_model(self):
        return self.model

    def decide(self):
        try:
            data = flask.request.json
            return flask.jsonify(dict(
                argument=data['argument'],
                setting = self.setting,
                model=self.model
            ))
        except:
            return traceback.format_exc(), 500

    def resources(self):
        try:
            folder_contents = {}
            path = '/resources'
            for root, _, files in os.walk(path):
                for file_name in files:
                    file_path = os.path.join(root, file_name)
                    relative_path = os.path.relpath(file_path, path)
                    try:
                        with open(file_path, 'r', encoding='utf-8') as file:
                            folder_contents[relative_path] = file.read()
                    except:
                        folder_contents[relative_path] = None
            return folder_contents
        except:
            return traceback.format_exc(), 500


from typing import Iterable
from ....framework import DockerWebServiceApi, FileLike, File
import requests
from .settings import ZonosSettings
from .controller import ZonosController
import shutil
import os


class Zonos(DockerWebServiceApi[ZonosSettings, ZonosController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def train(self, speaker: str, samples: Iterable[FileLike.Type]):
        folder = self.controller.resource_folder('voices',speaker)
        shutil.rmtree(folder, ignore_errors=True)
        os.makedirs(folder)
        for sample in samples:
            with open(folder/FileLike.get_name(sample), 'wb') as stream:
                with FileLike(sample, self.cache_folder) as data:
                    stream.write(data.read())
        result = requests.post(
            self.endpoint(f'/train/{speaker}')
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint train returned error\n{result.text}")

    def voiceover(self, text: str, speaker: str, language='en-us'):
        result = requests.post(
            self.endpoint('/voiceover'),
            json = dict(
                text=text,
                speaker=speaker,
                language=language
            )
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint voiceover returned error\n{result.text}")
        return File(
            self.current_job_id+'.output.wav',
            result.content
        )


    Settings = ZonosSettings
    Controller = ZonosController



from dataclasses import dataclass
from ....framework import ConnectionSettings

@dataclass
class HelloBrainBoxSettings:
    connection = ConnectionSettings(20000, 5)
    setting: str = 'default_setting'



import copy

from ...deployment import LocalExecutor, IExecutor, Command, IFileSystem, Machine
from ...common import Logger
from dataclasses import dataclass
from datetime import datetime
from enum import Enum



class Log:
    @dataclass
    class Item:
        timestamp: datetime
        data: str

    def __init__(self):
        self.items: list['Log.Item'] = []


class LogLogger(Logger):
    def __init__(self, log: Log):
        self._log = log

    def report_progress(self, progress: float):
        self.log(f"Progress report: {100*progress}%")

    def log(self, s):
        self._log.items.append(Log.Item(
            datetime.now(),
            s,
        ))





class LoggingLocalExecutor(IExecutor):
    def __init__(self, log: Log):
        self.base_executor = LocalExecutor()
        self.log = log

    def add_item(self, s):
        print(s)
        self.log.items.append(Log.Item(datetime.now(), '< ' + str(s)))

    def execute_command(self, command: Command):
        self.log.items.append(Log.Item(
            datetime.now(),
            '> ' + ' '.join(command.command)
        ))
        command = copy.copy(command)
        if command.options is None:
            command.options = Command.Options()
        command.options.monitor_output = self.add_item

        result = self.base_executor.execute_command(command)
        return result

    def get_fs(self) -> IFileSystem:
        return self.base_executor.get_fs()

    def get_machine(self) -> Machine:
        return self.base_executor.get_machine()




from ...common import Logger
from ..core import OperatorMessage
from queue import Queue

class DeciderLogger(Logger):
    def __init__(self, id: str, queue: Queue[OperatorMessage]):
        self.id = id
        self.queue = queue

    def report_progress(self, progress: float):
        self.queue.put(OperatorMessage(self.id, OperatorMessage.Type.report_progress, progress))

    def log(self, s):
        self.queue.put(OperatorMessage(self.id, OperatorMessage.Type.log, s))



from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import VoskModel

@dataclass
class VoskSettings:
    connection = ConnectionSettings(20105, 20)

    models_to_download = (
        VoskModel(
            'en',
            'https://alphacephei.com/vosk/models/vosk-model-en-us-0.22-lgraph.zip'
        ),
    )



from typing import *
from abc import ABC, abstractmethod
from ...job_processing import Job
from uuid import uuid4

class IBrainBoxTask(ABC):
    @abstractmethod
    def create_jobs(self) -> list[Job]:
        pass

    @abstractmethod
    def get_resulting_id(self) -> str|None:
        pass

    def postprocess_result(self, result, api):
        return result

    def before_add(self, api):
        pass

    @staticmethod
    def to_all_jobs(tasks: Iterable['IBrainBoxTask']) -> list[Job]:
        jobs = []
        for task in tasks:
            jobs.extend(task.create_jobs())
        return jobs

    @staticmethod
    def safe_id():
        return 'id_'+str(uuid4()).replace('-','')


from brainbox.deciders import HelloBrainBox

if __name__ == '__main__':
    controller = HelloBrainBox.Controller()
    controller.install()
    controller.self_test()
    #controller.run_notebook()




from abc import ABC, abstractmethod

class ISingleLoadableModelApi(ABC):
    @abstractmethod
    def load_model(self, model_name: str):
        pass

    @abstractmethod
    def get_loaded_model_name(self) -> str|None:
        pass



from pydantic import BaseModel
from typing import Dict, List, Any

class CoordsFaceSquare(BaseModel):
    objects: List[Dict]

class ImageFormat(BaseModel):
    image_base64: str

class LoadModelRequest(BaseModel):
    model_id: str



import os
from ....framework import DockerWebServiceApi, File, FileLike, BrainBoxApi
from .workflows import IWorkflow, TextToImage, Upscale, WD14Interrogate
from .settings import ComfyUISettings
from .controller import ComfyUIController
import requests
import time
from pathlib import Path
from yo_fluq import FileIO


class ComfyUI(DockerWebServiceApi[ComfyUISettings, ComfyUIController]):
    def __init__(self, address: str = None):
        super().__init__(address)


    def run_on_dictionary(self, workflow_type:type|str, **kwargs):
        if isinstance(workflow_type, str):
            workflow_type = getattr(ComfyUI.Workflows, workflow_type)
        workflow = workflow_type(**kwargs)
        return self.run_workflow(workflow)

    def __call__(self, workflow: IWorkflow):
        self.run_workflow(workflow)


    def run_workflow(self, workflow: IWorkflow, **kwargs):
        for key, value in kwargs.items():
            setattr(workflow, key, value)

        input_files = []

        for file in workflow.get_input_files():
            input_file_name = self.controller.resource_folder('input') / FileLike.get_name(file)
            input_files.append(input_file_name)
            with FileLike(file, self.cache_folder) as content:
                FileIO.write_bytes(content.read(), input_file_name)

        json = workflow.create_workflow_json(self.current_job_id)

        address = f'http://{self.address}/prompt'
        result = requests.post(address, json=dict(prompt=json))
        if result.status_code!=200:
            raise ValueError(
                f"Decider ComfyUI could not schedule the workflow {type(workflow)}:\n\n{result.text}\n\n{json}\n\n{[f.name for f in input_files]}"
            )

        while True:
            queue_state = requests.get(address).json()
            remaining = queue_state.get('exec_info', {}).get('queue_remaining', None)
            if remaining is None:
                raise ValueError(f"Unexpected state reply {queue_state}")
            if remaining == 0:
                break
            time.sleep(0.1)

        resulting_files = []
        output_folder = self.controller.resource_folder('output')
        for file in sorted(os.listdir(output_folder)):
            if file.startswith(self.current_job_id):
                resulting_files.append(File.read(output_folder/file))
                os.unlink(output_folder/file)

        for input_file_name in input_files:
            os.unlink(input_file_name)

        return workflow.postprocess(resulting_files)


    @staticmethod
    def get_models(api: BrainBoxApi):
        return api.controller_api.list_resources(ComfyUI, 'models/')

    Controller = ComfyUIController
    Settings = ComfyUISettings

    class Workflows:
        Upscale = Upscale
        TextToImage = TextToImage
        WD14Interrogate = WD14Interrogate














from pathlib import Path
from typing import cast
from ....framework import IController, DownloadableModel
from dataclasses import dataclass

@dataclass
class WhisperModel(DownloadableModel):
    name: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder() / (self.name + '.pt')

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import WhisperController
        controller = cast(WhisperController, controller)
        instance_id = controller.run(None)
        api = controller.find_api(instance_id)
        for model in models:
            model = cast(WhisperModel, model)
            api.load_model(model.name)
        controller.stop(instance_id)







from dataclasses import dataclass

@dataclass
class TrainingSettings:
    language = 'en-us'
    sample_rate = 22050
    base_model = 'lessac.ckpt'
    batch_size: int = 8
    validation_split: float = 0.0
    max_epochs: int = 2170
    checkpoint_epochs: int = 1
    precision: int = 32
    num_test_examples: int = 0
    keep_intermediate: bool = False

    



import datetime
from brainbox.deciders import OpenTTS, Collector
from brainbox.framework import FileIO, Loc
from brainbox import MediaLibrary, BrainBoxTask, BrainBoxApi
from pathlib import Path
from yo_fluq import Query

texts = {
    'train': ['computer' ,'how are you','how is it going' ,'make me a sandwich','set the timer for ten minutes'],
    'test': ["what's the weather", "remind me to wash dishes"]
}

if __name__ == '__main__':
    tags = []
    for voice in ['p256','p257']:
        for split in texts:
            for text in texts[split]:
                tags.append(dict(speaker=voice, split=split, text=text))
    pack = (
        Query.en(tags)
        .feed(Collector.FunctionalTaskBuilder(
            lambda z: BrainBoxTask.call(OpenTTS).voiceover(z['text'], speakerId=z['speaker'])
        ))
    )

    with BrainBoxApi.Test() as api:
        lib = api.execute(pack)
        api.download(lib, Path(__file__).parent/'test_media_library.zip', True)




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, INotebookableController
)
from ...common import VOICEOVER_TEXT, check_if_its_sound
from .settings import CoquiTTSSettings
from pathlib import Path


class CoquiTTSController(DockerWebServiceController[CoquiTTSSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return CoquiTTSSettings()

    def create_api(self):
        from .api import CoquiTTS
        return CoquiTTS()

    def post_install(self):
        for model in self.settings.builtin_models_to_download:
            path = self.resource_folder('builtin') / model.due_folder_name
            if not path.is_dir():
                cfg = self.get_service_run_configuration(None).as_service_worker(
                    '--install',
                    model.model_name,
                    '--install-mode',
                    model.mode
                )
                self.run_auxiliary_configuration(cfg)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .tests import Test
        yield TestReport.attach_source_file(Test)
        yield from Test(api, tc).test_all(self.settings)






DOCKERFILE = f'''
FROM python:3.11

{SmallImageBuilder.APT_INSTALL('espeak')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

{SmallImageBuilder.GIT_CLONE_AND_RESET(
    'https://github.com/coqui-ai/TTS',
    'TTS',
    'dbf1a08a0d4e47fdad6172e433eeb34bc6b13b4e', 
    '[all,dev,notebooks]'
)}

COPY . /home/app/TTS/kaia

ENTRYPOINT ["python3","/home/app/TTS/kaia/main.py"]
'''

DEPENDENCIES = '''
absl-py==2.1.0
aiohappyeyeballs==2.4.4
aiohttp==3.11.11
aiosignal==1.3.2
annotated-types==0.7.0
anyascii==0.3.2
anyio==4.7.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
astroid==2.7.3
asttokens==3.0.0
async-lru==2.0.4
attrs==24.3.0
audioread==3.0.1
babel==2.16.0
bangla==0.0.2
beautifulsoup4==4.12.3
black==24.10.0
bleach==6.2.0
blinker==1.9.0
blis==1.1.0
bnnumerizer==0.0.2
bnunicodenormalizer==0.1.7
bokeh==1.4.0
catalogue==2.0.10
certifi==2024.12.14
cffi==1.17.1
charset-normalizer==3.4.0
click==8.1.7
cloudpathlib==0.20.0
comm==0.2.2
confection==0.1.5
contourpy==1.3.1
coqpit==0.0.17
coverage==7.6.9
cutlet==0.4.0
cycler==0.12.1
cymem==2.0.10
Cython==3.0.11
dateparser==1.1.8
debugpy==1.8.11
decorator==5.1.1
defusedxml==0.7.1
docopt==0.6.2
einops==0.8.0
encodec==0.1.1
executing==2.1.0
fastjsonschema==2.21.1
filelock==3.16.1
Flask==3.1.0
fonttools==4.55.3
fqdn==1.5.1
frozenlist==1.5.0
fsspec==2024.12.0
fugashi==1.4.0
g2pkk==0.1.2
grpcio==1.68.1
gruut==2.2.3
gruut-ipa==0.13.0
gruut-lang-de==2.0.1
gruut-lang-en==2.0.1
gruut-lang-es==2.0.1
gruut-lang-fr==2.0.2
h11==0.14.0
hangul-romanize==0.1.0
httpcore==1.0.7
httpx==0.28.1
huggingface-hub==0.27.0
idna==3.10
inflect==7.4.0
ipykernel==6.29.5
ipython==8.31.0
isoduration==20.11.0
isort==5.13.2
itsdangerous==2.2.0
jaconv==0.4.0
jamo==0.4.1
jedi==0.19.2
jieba==0.42.1
Jinja2==3.1.4
joblib==1.4.2
json5==0.10.0
jsonlines==1.2.0
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-events==0.11.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.15.0
jupyter_server_terminals==0.5.3
jupyterlab==4.3.4
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
kiwisolver==1.4.7
langcodes==3.5.0
language_data==1.3.0
lazy-object-proxy==1.10.0
lazy_loader==0.4
librosa==0.10.2.post1
llvmlite==0.43.0
marisa-trie==1.2.1
Markdown==3.7
markdown-it-py==3.0.0
MarkupSafe==3.0.2
matplotlib==3.10.0
matplotlib-inline==0.1.7
mccabe==0.6.1
mdurl==0.1.2
mecab-python3==1.0.6
mistune==3.0.2
mojimoji==0.0.13
more-itertools==10.5.0
mpmath==1.3.0
msgpack==1.1.0
multidict==6.1.0
murmurhash==1.0.11
mutagen==1.47.0
mypy-extensions==1.0.0
nbclient==0.10.2
nbconvert==7.16.4
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==2.8.8
nltk==3.9.1
nose2==0.15.1
notebook==7.3.2
notebook_shim==0.2.4
num2words==0.5.14
numba==0.60.0
numpy==1.26.4
nvidia-cublas-cu12==12.4.5.8
nvidia-cuda-cupti-cu12==12.4.127
nvidia-cuda-nvrtc-cu12==12.4.127
nvidia-cuda-runtime-cu12==12.4.127
nvidia-cudnn-cu12==9.1.0.70
nvidia-cufft-cu12==11.2.1.3
nvidia-curand-cu12==10.3.5.147
nvidia-cusolver-cu12==11.6.1.9
nvidia-cusparse-cu12==12.3.1.170
nvidia-nccl-cu12==2.21.5
nvidia-nvjitlink-cu12==12.4.127
nvidia-nvtx-cu12==12.4.127
overrides==7.7.0
packaging==24.2
pandas==1.5.3
pandocfilters==1.5.1
parso==0.8.4
pathspec==0.12.1
pexpect==4.9.0
pillow==11.0.0
platformdirs==4.3.6
pooch==1.8.2
preshed==3.0.9
prometheus_client==0.21.1
prompt_toolkit==3.0.48
propcache==0.2.1
protobuf==5.29.2
psutil==6.1.1
ptyprocess==0.7.0
pure_eval==0.2.3
pycparser==2.22
pydantic==2.10.4
pydantic_core==2.27.2
Pygments==2.18.0
pylint==2.10.2
pynndescent==0.5.13
pyparsing==3.2.0
pypinyin==0.53.0
pysbd==0.3.4
python-crfsuite==0.9.11
python-dateutil==2.9.0.post0
python-json-logger==3.2.1
pytz==2024.2
PyYAML==6.0.2
pyzmq==26.2.0
referencing==0.35.1
regex==2024.11.6
requests==2.32.3
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rich==13.9.4
rpds-py==0.22.3
safetensors==0.4.5
scikit-learn==1.6.0
scipy==1.14.1
Send2Trash==1.8.3
shellingham==1.5.4
six==1.17.0
smart-open==7.1.0
sniffio==1.3.1
soundfile==0.12.1
soupsieve==2.6
soxr==0.5.0.post1
spacy==3.8.3
spacy-legacy==3.0.12
spacy-loggers==1.0.5
srsly==2.5.0
stack-data==0.6.3
SudachiDict-core==20241021
SudachiPy==0.6.9
sympy==1.13.1
tensorboard==2.18.0
tensorboard-data-server==0.7.2
terminado==0.18.1
thinc==8.3.3
threadpoolctl==3.5.0
tinycss2==1.4.0
tokenizers==0.21.0
toml==0.10.2
torch==2.5.1
torchaudio==2.5.1
tornado==6.4.2
tqdm==4.67.1
trainer==0.0.36
traitlets==5.14.3
transformers==4.47.1
triton==3.1.0
typeguard==4.4.1
typer==0.15.1
types-python-dateutil==2.9.0.20241206
typing_extensions==4.12.2
tzlocal==5.2
umap-learn==0.5.7
Unidecode==1.3.8
unidic-lite==1.0.8
uri-template==1.3.0
urllib3==2.2.3
wasabi==1.1.3
wcwidth==0.2.13
weasel==0.4.1
webcolors==24.11.1
webencodings==0.5.1
websocket-client==1.8.0
Werkzeug==3.1.3
wrapt==1.12.1
yarl==1.18.3
'''


from .api import CoquiTTS
from ....framework import TestReport, BrainBox, File
from ...common import VOICEOVER_TEXT, check_if_its_sound
from unittest import TestCase
from pathlib import Path
from .settings import CoquiTTSSettings

TEST_VOICE_PATH = Path(__file__).parent / 'test_voice.wav'


class Test:
    def __init__(self, api: BrainBox.Api, tc: TestCase):
        self.api = api
        self.tc = tc

    def test_dub(self):
        file = self.api.open_file(self.api.execute(
            BrainBox.Task.call(CoquiTTS).dub(text=VOICEOVER_TEXT)
        ))
        yield TestReport.last_call(self.api).result_is_file(File.Kind.Audio).with_comment("Voiceover")
        check_if_its_sound(file.content, self.tc)



    def test_voice_clone(self):
        file = self.api.open_file(self.api.execute(
            BrainBox.Task.call(CoquiTTS).voice_clone(VOICEOVER_TEXT, voice='test_voice')
        ))
        yield (
            TestReport
            .last_call(self.api)
            .result_is_file(File.Kind.Audio)
            .with_uploaded_file('test_voice', File.read(TEST_VOICE_PATH))
            .with_comment("Voice clone")
        )
        check_if_its_sound(file.content, self.tc)

    def load_model_and_repr_description(self, model_name: str):
        yield TestReport.section(f"Model {model_name}")
        self.api.execute(BrainBox.Task.call(CoquiTTS).load_model(model_name))
        yield TestReport.last_call(self.api).with_comment("Loading model (will be used if no `model` provided in decider's call)")




    def test_all(self, settings: CoquiTTSSettings):
        CoquiTTS.upload_voice('test_voice', TEST_VOICE_PATH).execute(self.api)

        for model in settings.builtin_models_to_download:
            yield from self.load_model_and_repr_description(model.model_name)
            if model.test_dub:
                yield from self.test_dub()
            if model.test_voice_clone:
                yield from self.test_voice_clone()

        custom_models_files = self.api.controller_api.list_resources(CoquiTTS, 'custom')
        if custom_models_files is None:
            return
        print(custom_models_files)
        custom_models = [
            c.split('/')[-1]
            for c in custom_models_files
            if c.endswith('.pth') and not c.endswith('.speakers.pth')
        ]

        for custom_model in custom_models:
            yield from self.load_model_and_repr_description('custom/'+custom_model)
            yield from self.test_dub()








import pandas as pd

from ....job_processing import Job
from .batch_page import _select, _pretty_sec
from sqlalchemy import select, case, func
from yo_fluq import *
import numpy as np
from ....common import HTML

def _get_frames(session):
    cte = _select().cte('processed_jobs')

    status_query = (
        session.query(
            cte.c.batch,
            cte.c.status,
            func.count(cte.c.id).label("count"),
        ).group_by(cte.c.batch, cte.c.status)
    )

    stats_query = (
        session.query(
            cte.c.batch,
            func.min(cte.c.received_timestamp).label('received_timestamp'),
            func.max(cte.c.finished_timestamp).label('finished_timestamp'),
            func.avg(cte.c.in_work).label('processing_time'),
            func.sum(cte.c.in_work).label('total_processing_time'),
            func.avg(cte.c.progress).label('progress'),
        ).group_by(cte.c.batch)
    )

    names_query = (
        session.query(
            cte.c.batch,
            cte.c.decider,
            func.count(cte.c.id).label('amount')
        ).group_by(cte.c.batch, cte.c.decider)
    )

    df_status = pd.read_sql_query(status_query.statement, session.bind)
    df_stats = pd.read_sql_query(stats_query.statement, session.bind)
    df_names = pd.read_sql_query(names_query.statement, session.bind)

    return df_status, df_stats, df_names


def _get_merged_frame(df_status, df_stats, df_names):
    if df_status.shape[0] == 0:
        return None
    df = df_status
    df = df.pivot_table(index='batch', columns='status', values='count').fillna(0)
    for column in ['success','in_work','failure','waiting']:
        if column not in df.columns:
            df[column] = 0
    df['total'] = df.sum(axis=1)
    df['finished'] = df.success + df.failure
    df = df.astype(int)
    df.columns = ['tasks_' + c for c in df.columns]
    df['is_finished'] = df.tasks_finished == df.tasks_total

    df = df_stats.merge(df, right_index=True, left_on='batch')

    df_names['part'] = df_names['decider'] + ' x ' + df_names['amount'].astype(str) + "\n"
    names = df_names.groupby('batch').part.sum()
    df = df.merge(names.to_frame('names'), left_on='batch', right_index=True)

    df = df.sort_values('received_timestamp', ascending=False)

    pd.options.display.width = None
    return df

def _to_html(df):
    html = []

    for row in Query.df(df.reset_index()):
        html.append('<tr>')
        html.append(f'<td><a href="/html/jobs/batch_page/{row.batch}">{row.batch}</td>')
        names = "<br>".join([z for z in row.names.split('\n') if z!=''])
        html.append(f'<td>{names}</td>')
        eta = None
        if row.tasks_total == 1:
            progress = row.progress
            if progress is not None and progress > 0:
                eta = row.processing_time * (1 - progress) / progress
        else:
            progress = row.tasks_finished / row.tasks_total
            if row.processing_time is not None:
                eta = row.processing_time * (row.tasks_total - row.tasks_finished)

        html.append(f'<td>{row.received_timestamp}</td>')

        html.append('<td>')
        if row.is_finished:
            html.append(str(row.finished_timestamp))
            if row.total_processing_time is not None:
                html.append("<br>")
                html.append(_pretty_sec(row.total_processing_time))
        else:
            if progress is not None and not np.isnan(progress):
                html.append(f'{int(100 * progress)}%')
                if eta is not None:
                    html.append(f'<br>{_pretty_sec(eta)}')
            else:
                html.append('?')
        html.append("</td>")

        html.append("<td>")
        html.append(HTML.button("/jobs/cancel", "Cancel", dict(batch_id=row.batch)))
        html.append("</td>")




    return '\n'.join(html)


def create_main_page(session):
    df = _get_merged_frame(*_get_frames(session))
    if df is None:
        html = ''
    else:
        html = _to_html(df)

    html = '''
[<a href="/html/jobs/operator_log">Operator log</a>]&nbsp;[<a href="/html/controllers/status">Controllers</a>]</br>
<table border=1 cellspacing=0>
    ''' + html + "</table>"

    return HTML.page("Brainbox: Batches", html)


import sys
import subprocess
from server import WhisperApp
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Small web server around Whisper model"
    )
    parser.add_argument('-n', '--notebook', action='store_true')

    args = parser.parse_args()
    print(f'Running with arguments {args}')

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0', "--NotebookApp.token=''"], cwd='/repo')
        exit(0)

    factory = WhisperApp()
    factory.create_app().run('0.0.0.0', 8084)




import os
import traceback
import uuid
from converter import Converter
import traceback
from pathlib import Path
import pickle
from pydantic import BaseModel
import flask


OUTPUT_DIR = '/home/app/outputs'


class Server:
    def __init__(self):
        self.converter = Converter()
        self.loaded_models = {}

    def __call__(self):
        app = flask.Flask("OpenVoiceServer")
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.add_url_rule('/train/<model>', view_func=self.train, methods=['POST'])
        app.add_url_rule('/generate/<model>', view_func=self.generate, methods=['POST'])
        app.add_url_rule('/generate/<model>/<source_model>', view_func=self.generate, methods=['POST'])
        app.run('0.0.0.0', port=8080)



    def index(self):
        return "OK"

    def train(self, model: str):
        try:
            from train import train
            train(model, self.converter)
            return "OK"
        except:
            result = traceback.format_exc()
            return result, 500

    def _get_model(self, model):
        if model not in self.loaded_models:
            model_path = Path(f'/models/{model}')
            if not model_path.is_file():
                raise ValueError(f"Model {model} was not trained")
            with open(f'/models/{model}', 'rb') as stream:
                self.loaded_models[model] = pickle.load(stream)
        return self.loaded_models[model]

    def generate(self, model, source_model=None):
        source_path = None
        output_path = None
        try:
            id = uuid.uuid4()
            output_path = f"{OUTPUT_DIR}/output_{id}.wav"
            source_path = f"{OUTPUT_DIR}/source_{id}.wav"

            os.makedirs(OUTPUT_DIR, exist_ok=True)
            file = flask.request.files['file']
            file.save(source_path)

            target_se = self._get_model(model)
            if source_model is not None:
                source_se = self._get_model(source_model)
            else:
                source_se = self.converter.convert(source_path)

            encode_message = "@MyShell"

            self.converter.tone_color_converter.convert(
                audio_src_path=source_path,
                src_se=source_se,
                tgt_se=target_se,
                output_path=output_path,
                message=encode_message
            )

            if not os.path.exists(output_path):
                raise ValueError("OpenVoice failed to create an output file")

            return flask.send_file(output_path)
        except:
            return traceback.format_exc(), 500
        finally:
            for path in [source_path, output_path]:
                if path is not None and os.path.exists(path):
                    os.remove(path)



import os
from pathlib import Path
import shutil
import subprocess
from settings import TrainingSettings

class Functions:
    def __init__(self, dataset: str, working_dir: str, settings: TrainingSettings):
        self.dataset = dataset
        self.working_dir = Path(working_dir)
        self.settings = settings


    def get_path(self):
        return self.working_dir/'trainings'/self.dataset

    def is_single_speaker(self):
        with open(self.get_path()/ 'source/metadata.csv', 'r') as file:
            first_line = file.readline()
            groups = first_line.split('|')
            if len(groups) == 2:
                return True
            elif len(groups) == 3:
                return False
            else:
                raise ValueError(
                    f"Lines of csv files are supposed to have 2 or 3 parts separated by '|', but was\n{first_line}")


    def preprocess(self):
        source_path = self.get_path()/'source'
        preprocessed_path = self.get_path()/'preprocessed'
        single_speaker = self.is_single_speaker()

        shutil.rmtree(preprocessed_path, ignore_errors=True)

        subprocess.call([
            'python3',
            '-m',
            'piper_train.preprocess',
            '--language',
            self.settings.language,
            '--input-dir',
            str(source_path),
            '--output-dir',
            str(preprocessed_path),
            '--dataset-format',
            'ljspeech',
            *(['--single-speaker'] if single_speaker else []),
            '--sample-rate',
            str(self.settings.sample_rate)
        ])

    def train(self):
        preprocessed_path = self.get_path()/'preprocessed'

        subprocess.call([
            'python3',
            '-m',
            'piper_train',
            '--dataset-dir',
            str(preprocessed_path),
            '--accelerator',
            "cuda",
            '--devices',
            '1',
            '--batch-size',
            str(self.settings.batch_size),
            '--validation-split',
            str(self.settings.validation_split),
            '--num-test-examples',
            str(self.settings.num_test_examples),
            '--max_epochs',
            str(self.settings.max_epochs),
            '--resume_from_checkpoint',
            str(self.working_dir/'base_models'/self.settings.base_model),
            '--checkpoint-epochs',
            str(self.settings.checkpoint_epochs),
            '--precision',
            str(self.settings.precision),
            *(['--keep-intermediate'] if self.settings.keep_intermediate else []),
        ])

    def export_model(self):
        preprocessed_path = self.get_path()/'preprocessed'
        model_path = preprocessed_path/'lightning_logs/version_0/checkpoints/'
        result_path = self.get_path()/'result'
        os.makedirs(result_path, exist_ok=True)
        for model in os.listdir(model_path):
            name, extension = model.split('.')
            Functions.convert_custom_filename(model_path/model, result_path)
            shutil.copy(
                preprocessed_path/'config.json',
                str(result_path/(name+'.onnx.json'))
            )

    @staticmethod
    def convert_custom_filename(path, target_folder = None):
        if target_folder is None:
            target_folder = path.parent
        name, extension = path.name.split('.')
        subprocess.call([
            'python3',
            '-m',
            'piper_train.export_onnx',
            str(path),
            str(target_folder / (name + ".onnx"))
        ])



from ....framework import DockerWebServiceApi, FileLike
import requests
from .settings import VoskSettings
from .controller import VoskController
from .model import VoskModel


class Vosk(DockerWebServiceApi[VoskSettings, VoskController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def load_model(self, model_name):
        result = requests.post(self.endpoint('load_model'), json=dict(model_name=model_name))
        if result.status_code != 200:
            raise ValueError(f"Endpoint load_model returned exception\n{result.text}")

    def transcribe(self, file: FileLike.Type, model_name: str|None = None):
        with FileLike(file, self.cache_folder) as stream:
            ep = 'transcribe'
            if model_name is not None:
                ep+='/'+model_name
            result = requests.post(
                self.endpoint('transcribe/'+model_name),
                files=(
                    ('file', stream),
                ))
            if result.status_code != 200:
                raise ValueError(f"Endpoint transcribe returned error\n{result.text}")
            return result.json()

    Settings = VoskSettings
    Controller = VoskController
    Model = VoskModel



import os
import shutil
import sys
from pathlib import Path
from dataclasses import dataclass
from train import Trainer
from utils import train_path, rhasspy_path
from transcribe import Transcriber


@dataclass
class Model:
    name: str

    def get_path(self):
        return train_path/self.name


    def train(self, base_language: str, sentences: str, custom_words: str):
        shutil.rmtree(self.get_path(), True)
        shutil.copytree(rhasspy_path/base_language, self.get_path())
        trainer = Trainer(self.get_path(), sentences, base_language, custom_words)
        return trainer.train()

    def create_transcriber(self) -> Transcriber:
        return Transcriber(self.get_path())










from ...framework import IDecider
import string
import random
import time

class FakeText(IDecider):
    def __call__(self, prefix = '', length = 50, time_to_sleep: float|None = None):
        if time_to_sleep is not None:
            time.sleep(time_to_sleep)
        characters = string.ascii_letters + string.digits + string.punctuation
        random_text = prefix+' '+''.join(random.choices(characters, k=length))
        return random_text



import numpy as np
from PIL import Image
from frame import Frame
from typing import Iterable


def layer_comparator(iterable: Iterable[Frame]):
    old_array = None


    for frame in iterable:
        array = np.array(frame.pil_image)

        if old_array is not None:
            frame.simple_comparator_delta = np.mean(np.abs(array - old_array))

        old_array = array


from .hello_brainbox import HelloBrainBox
from .collector import Collector
from .fake_file import FakeFile
from .fake_text import FakeText
from .boilerplate_server import BoilerplateServer
from .boilerplate_on_demand import BoilerplateOnDemand
from .cache_manager import CacheManager
from .ffmpeg import Ffmpeg
from .empty_task import EmptyTask


from dataclasses import dataclass
from .model import PiperTrainingModel

@dataclass
class PiperTrainingSettings:
    models_to_download = (
        PiperTrainingModel(
            'lessac.ckpt',
            'https://huggingface.co/datasets/rhasspy/piper-checkpoints/resolve/main/en/en_US/lessac/medium/epoch%3D2164-step%3D1355540.ckpt?download=true'
        )
    )


from ..core import ICoreAction, Job, Core
from sqlalchemy.orm import Session
from sqlalchemy import select
from datetime import datetime


class CheckReadyAction(ICoreAction):
    def apply(self, core: Core):
        self.assign_ready_to_independent_tasks(core)
        self.assign_ready_to_dependent_tasks(core)

    def assign_ready_to_independent_tasks(self, core: Core):
        with core.new_session() as session:
            (
                session
                .query(Job)
                .filter(
                    ~Job.finished & ~Job.ready & ~Job.has_dependencies
                )
                .update({Job.ready: True, Job.ready_timestamp:datetime.now()})
            )
            session.commit()


    def assign_ready_to_dependent_tasks(self, core: Core):
        with core.new_session() as session:
            dependent_tasks = list(session.execute(
                select(Job.id, Job.dependencies)
                .where(~Job.finished & ~Job.ready & Job.has_dependencies)
            ))

            dependency_ids = list(set(id for dependent in dependent_tasks for id in dependent.dependencies.values()))

            dependency_status = list(session.execute(
                select(Job.id, Job.finished)
                .where(Job.id.in_(dependency_ids))
            ))
            id_to_finished = {status.id: status.finished for status in dependency_status}

            for task in dependent_tasks:
                set_ready = True
                for id in task.dependencies.values():
                    if id not in id_to_finished:
                        core.close_job(
                            session,
                            core.get_job_by_id(session, task.id),
                            f"id {id} is required by this task but is absent"
                        )
                    elif not id_to_finished[id]:
                        set_ready = False
                        break
                if set_ready:
                    task_obj = core.get_job_by_id(session, task.id)
                    task_obj.ready = True
                    task_obj.ready_timestamp = datetime.now()
                    core.operator_log.task(task_obj.id).event('Ready')


            session.commit()


import requests

from ....framework import (
    ResourcePrerequisite, DockerWebServiceApi, File, CombinedPrerequisite,
    ISingleLoadableModelApi, IPrerequisite, BrainBoxApi
)
from .controller import CoquiTTSController
from .settings import CoquiTTSSettings
from pathlib import Path

class CoquiTTS(DockerWebServiceApi[CoquiTTSSettings, CoquiTTSController], ISingleLoadableModelApi):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def load_model(self, model: str):
        result = requests.post(
            f'http://{self.address}/load_model',
            json=dict(model=model)
        )
        if result.status_code != 200:
            raise ValueError(f"Error when calling `load_model`\n{result.text}")
        return result.json()

    def get_loaded_model(self):
        result =requests.get(
            self.endpoint('/get_loaded_model')
        )
        if result.status_code!=200:
            raise ValueError(f'Error when calling `get_loaded_model`\n{result.text}')
        return result.json()

    def get_loaded_model_name(self) -> str|None:
        model = self.get_loaded_model()
        if model is None:
            return None
        return model['name']

    def dub(self, text: str, model: str|None = None, voice: str|None = None, language: str|None = None):
        reply = requests.post(
            self.endpoint('/dub'),
            json=dict(text=text, model=model, voice=voice, language=language)
        )
        if reply.status_code != 200:
            raise ValueError(f"Error when calling `dub`\n{reply.text}")
        return File(self.current_job_id + '.output.wav', reply.content, File.Kind.Audio)

    def voice_clone(self, text: str, model: str|None = None, voice: str|None = None, language: str|None = None):
        reply = requests.post(
            self.endpoint('/voice_clone'),
            json=dict(text=text, model=model, voice=voice, language=language)
        )
        if reply.status_code != 200:
            raise ValueError(f"Error when calling `voice_clone`\n{reply.text}")
        return File(self.current_job_id + '.output.wav', reply.content, File.Kind.Audio)


    @staticmethod
    def export_model_from_training(source_model_file_path: Path, model_name: str) -> CombinedPrerequisite:
        target_path = f'custom/{model_name}.pth'
        command = []
        command.append(ResourcePrerequisite(
            CoquiTTS,
            target_path,
            source_model_file_path
        ))
        command.append(ResourcePrerequisite(
            CoquiTTS,
            target_path+'.json',
            source_model_file_path.parent/'config.json'
        ))
        command.append(ResourcePrerequisite(
            CoquiTTS,
            target_path+'.speakers.pth',
            source_model_file_path.parent.parent.parent / 'data/speakers.pth'
        ))
        return CombinedPrerequisite(command)


    @staticmethod
    def upload_voice(voice: str, *files) -> 'VoiceUploadPrerequisite':
        return VoiceUploadPrerequisite(voice, files)

    Controller = CoquiTTSController
    Settings = CoquiTTSSettings

    @classmethod
    def get_ordering_arguments_sequence(cls) -> tuple[str,...]|None:
        return ('model',)



class VoiceUploadPrerequisite(IPrerequisite):
    def __init__(self, voice: str, files: tuple[Path,...]):
        self.voice = voice
        self.files = files

    def execute(self, api: BrainBoxApi):
        ResourcePrerequisite(CoquiTTS,f'voices/{self.voice}', None).execute(api)
        for file in self.files:
            ResourcePrerequisite(CoquiTTS, f'voices/{self.voice}/{file.name}', file).execute(api)




import requests
from typing import Optional
from .controller import PiperController
from .settings import PiperSettings
from .model import PiperModel
from ....framework import DockerWebServiceApi, File, IPrerequisite, BrainBoxApi
from pathlib import Path


class Piper(DockerWebServiceApi[PiperSettings, PiperController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def voiceover(self,
                  text: str,
                  model: str = "en"
                  ) -> File:
       
        url = f"http://{self.address}/synthesize"
        payload = {
            "text": text,
            "model": model
        }
        reply = requests.post(url, json=payload, timeout=30)
        if reply.status_code != 200:
            raise ValueError(f"{reply.status_code}\n{reply.text}")
        
        return File(self.current_job_id+'.output.wav', reply.content, File.Kind.Audio)

    Controller = PiperController
    Settings = PiperSettings
    Model = PiperModel


    class UploadVoice(IPrerequisite):
        def __init__(self,
                     onnx_file_location: Path,
                     custom_json_file_location: Path|None = None,
                     custom_name: str|None = None
                     ):
            self.onnx_file_location = onnx_file_location
            self.custom_json_file_location = custom_json_file_location
            self.custom_name = custom_name

        def execute(self, api: BrainBoxApi):
            name = self.custom_name
            if name is None:
                name = self.onnx_file_location.name.split('.')[0]

            api.controller_api.upload_resource(
                'Piper',
                f'models/{name}.onnx',
                self.onnx_file_location
            )
            json_location = self.custom_json_file_location
            if json_location is None:
                json_location = self.onnx_file_location.parent/(self.onnx_file_location.name+'.json')
            api.controller_api.upload_resource(
                'Piper',
                f'models/{name}.onnx.json',
                json_location
            )

    class DeleteVoice(IPrerequisite):
        def __init__(self, name: str):
            self.name = name

        def execute(self, api: BrainBoxApi):
            api.controller_api.delete_resource('Piper', f'models/{self.name}.onnx')
            api.controller_api.delete_resource('Piper', f'models/{self.name}.onnx.json')



from functools import wraps
from dataclasses import dataclass


@dataclass
class EndpointMetadata:
    url: str|None
    method: str


def endpoint(
        *,
        url: str|None = None,
        method = 'POST',
    ):
    def decorator(func):
        # Add the metadata as an attribute of the function
        func._metadata = EndpointMetadata(url, method)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return decorator


import argparse
import json

from functions import Functions
from settings import TrainingSettings
from pathlib import Path

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset')
    parser.add_argument('-c', '--convert')
    parser.add_argument('-f', '--working-folder', default='/resources')

    args = parser.parse_args()
    print(f"Running with arguments\n{args}")

    if args.dataset is not None:
        settings_path = Path(args.working_folder)/'trainings'/args.dataset/'settings.json'
        with open(settings_path,'r') as stream:
            settings_js = json.load(stream)
        settings = TrainingSettings(**settings_js)
        print("Training settings:")
        print(settings)

        f = Functions(args.dataset, args.working_folder, settings)
        print(f'DATASET {f.dataset}')
        print('PREPROCESSING')
        f.preprocess()
        print('TRAINING')
        f.train()
        print('EXPORTING')
        f.export_model()

    if args.convert is not None:
        Functions.convert_custom_filename(Path(args.working_folder)/'conversions'/args.convert)









import copy


from ....framework import IPostprocessor, BrainBox, FileIO
from pathlib import Path

class LorasExporter(IPostprocessor):
    def __init__(self,
                 target_folder: Path,
                 only_last_version: bool = False,
                 cleanup: bool = True,
                 ):
        self.target_folder = Path(target_folder)
        self.only_last_version = only_last_version
        self.cleanup = cleanup

    def postprocess(self, result, api: BrainBox.Api):
        from .api import KohyaSS
        produced_files = api.controller_api.list_resources(KohyaSS, result['training_folder']+'/model')

        general_description = result['description']
        if general_description is None:
            general_description = {}

        for file in produced_files:
            name = file.split('/')[-1]
            prefix = result['training_id']
            suffix = '.safetensors'
            if not name.startswith(prefix):
                continue
            if not name.endswith(suffix):
                continue
            if self.only_last_version and name != prefix+suffix:
                continue
            api.controller_api.download_resource(KohyaSS, file, self.target_folder/name)
            description_filename = name+'.description.json'
            description = copy.deepcopy(general_description)
            description['interfix'] = name[len(prefix):-len(suffix)]
            description['model_filename'] = name
            FileIO.write_json(
                description,
                self.target_folder/description_filename
            )


        if self.cleanup:
            api.controller_api.delete_resource(KohyaSS, result['training_folder'], True)








from dataclasses import dataclass

@dataclass
class AnalysisSettings:
    source_file_name: str
    max_produced_frames: int|None = None
    buffer_by_laplacian_in_ms: int|None = None
    add_simple_comparator: bool = False
    add_semantic_comparator: bool = False


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, TestReport
)
from ...common import download_file
from .settings import ComfyUISettings, ComfyUIFile
from pathlib import Path
import os


class ComfyUIController(DockerWebServiceController[ComfyUISettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        main_dependencies = [c.strip() for c in DEPENDENCIES.split('\n') if c.strip()!='']
        extension_dependencies = list(self.settings.extension_requirements)

        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            [main_dependencies, extension_dependencies]
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")

        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8188},
            mount_resource_folders={
                'hf': '/home/app/.cache/huggingface',
                'output': '/home/app/ComfyUI/output',
                'input': '/home/app/ComfyUI/input',
                'custom_nodes': '/home/app/ComfyUI/custom_nodes',
                'models': '/home/app/ComfyUI/models'
            }
        )


    def get_default_settings(self):
        return ComfyUISettings()

    def create_api(self):
        from .api import ComfyUI
        return ComfyUI()

    def download_file(self, file_desc: ComfyUIFile, override: bool = False):
        if file_desc.models_subfolder is not None:
            folder = self.resource_folder('models') / file_desc.models_subfolder
        elif file_desc.custom_subfolder is not None:
            folder = self.resource_folder(file_desc.custom_subfolder)
        else:
            raise ValueError("Both models_subfolder and custom_subfolder are null")
        file = folder / file_desc.get_filename()
        if file.is_file():
            if not override:
                return
            else:
                os.unlink(file)
        download_file(file_desc.url, str(file))


    def post_install(self):
        self.run_auxiliary_configuration(self.get_service_run_configuration(None).as_service_worker('--fix'))
        for extension in self.settings.extensions:
            arguments = ['--install', extension.git_url]
            if extension.commit is not None:
                arguments.extend(['--commit', extension.commit])
            self.run_auxiliary_configuration(self.get_service_run_configuration(None).as_service_worker(*arguments))
        for model in self.settings.models_to_download:
            self.download_file(model)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .tests import Test
        yield TestReport.attach_source_file(Test)
        yield from Test(api, tc).test_all()







DOCKERFILE = f'''
FROM python:3.12

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

{SmallImageBuilder.GIT_CLONE_AND_RESET('https://github.com/comfyanonymous/ComfyUI', 'ComfyUI', '669d9e4c67849e380044871788eb2be615a50396', install=False)}

RUN mkdir /home/app/fix && cp -r /home/app/ComfyUI/models /home/app/fix/models && cp -r /home/app/ComfyUI/input /home/app/fix/input

COPY main.py /home/app/main.py

ENTRYPOINT ["python3","/home/app/main.py"]
'''

DEPENDENCIES = '''
--find-links https://download.pytorch.org/whl/torch_stable.html
--extra-index-url https://download.pytorch.org/whl/cu124
aiohappyeyeballs==2.4.3
aiohttp==3.10.10
aiosignal==1.3.1
attrs==24.2.0
certifi==2024.8.30
cffi==1.17.1
charset-normalizer==3.4.0
einops==0.8.0
filelock==3.16.1
frozenlist==1.5.0
fsspec==2024.10.0
huggingface-hub==0.26.1
idna==3.10
Jinja2==3.1.4
kornia==0.7.3
kornia_rs==0.1.5
MarkupSafe==3.0.2
mpmath==1.3.0
multidict==6.1.0
networkx==3.4.2
numpy==2.1.2
nvidia-cublas-cu12==12.4.5.8
nvidia-cuda-cupti-cu12==12.4.127
nvidia-cuda-nvrtc-cu12==12.4.127
nvidia-cuda-runtime-cu12==12.4.127
nvidia-cudnn-cu12==9.1.0.70
nvidia-cufft-cu12==11.2.1.3
nvidia-curand-cu12==10.3.5.147
nvidia-cusolver-cu12==11.6.1.9
nvidia-cusparse-cu12==12.3.1.170
nvidia-nccl-cu12==2.21.5
nvidia-nvjitlink-cu12==12.4.127
nvidia-nvtx-cu12==12.4.127
packaging==24.1
pillow==11.0.0
propcache==0.2.0
psutil==6.1.0
pycparser==2.22
PyYAML==6.0.2
regex==2024.9.11
requests==2.32.3
safetensors==0.4.5
scipy==1.14.1
sentencepiece==0.2.0
setuptools==75.2.0
soundfile==0.12.1
spandrel==0.4.0
sympy==1.13.1
tokenizers==0.20.1
torch==2.5.0+cu124
torchaudio==2.5.0+cu124
torchsde==0.2.6
torchvision==0.20.0+cu124
tqdm==4.66.5
trampoline==0.1.2
transformers==4.46.0
triton==3.1.0
typing_extensions==4.12.2
urllib3==2.2.3
yarl==1.16.0
'''


import json
import re

from .test_report import TestReport
from .test_report_item import TestReportItem
from ....common import File, HTML
import base64


class Builder:
    def __init__(self):
        self.html = []

    def replace_code(self, code):
        return HTML.escape_code(code)


    def add_code(self, code, style=''):
        code = self.replace_code(code)
        self.html.append(f'<p style="font-family: monospace; {style}">{code}</p>')

    def error_color_style(self):
        return 'color: #dd0000;'

    def add_file(self, file_content: File):
        if not isinstance(file_content, File):
            self.html.append(f"<p {self.error_color_style()}>Expected File, but was:</p>")
            self.add_code(str(file_content), self.error_color_style())
            return
        if file_content.kind == File.Kind.Json or file_content.kind == File.Kind.Text:
            self.add_code(file_content.string_content)
        elif file_content.kind == File.Kind.Audio:
            b64 = base64.b64encode(file_content.content).decode("utf-8")
            html=f'''
            <audio controls>
                <source src="data:audio/mpeg;base64,{b64}" type="audio/mpeg">
                    <p {self.error_color_style()}>Your browser does not support the audio element.</p>
            </audio>'''
            self.html.append(html)
        elif file_content.kind == File.Kind.Image:
            b64 = base64.b64encode(file_content.content).decode("utf-8")
            img_tag = f'<img src="data:;base64,{b64}">'
            self.html.append(img_tag)
        else:
            self.html.append(f'<p {self.error_color_style()}>File.Kind {file_content.kind} is not yet supported</p>')

    def add_files_dict(self, header, d: dict[str, File]):
        if len(d) > 0:
            self.html.append(f"<h3>{header}</h3>")
            for key, value in d.items():
                self.html.append(f"<h4>{key}</h4>")
                self.add_file(value)

    def process_item(self, item: TestReportItem):
        self.html.append('<h2')
        if item.href is not None:
            self.html.append(f' id="{item.href}"')
        self.html.append('>')

        self.html.append(item.decider)
        if item.method is not None:
            self.html.append(f"::{item.method}")
        if item.parameter is not None:
            self.html.append(f' ({item.parameter})')
        self.html.append("</h2>")

        if item.finished_timestamp is not None and item.accepted_timestamp is not None:
            self.html.append("<p>")
            self.html.append( str( (item.finished_timestamp - item.accepted_timestamp).total_seconds()) +" sec")
            self.html.append("</p>")

        if item.comment is not None:
            self.html.append("<p>")
            self.html.append(item.comment)
            self.html.append("</p>")
        self.html.append("<h3>Arguments</h3>")
        for key, value in item.arguments.items():
            self.html.append(f'<h4>{key}</h4>')
            if isinstance(value, File):
                self.html.append("<i>File</i><br>")
                self.add_file(value)
            elif isinstance(value, list) or isinstance(value, dict):
                try:
                    self.add_code(json.dumps(value, indent=2))
                except:
                    self.html.append("<b>cannot render</b>")
            else:
                self.add_code(str(value))

        self.add_files_dict("Uploaded files", item.uploaded_files)
        self.add_files_dict("Used resources", item.used_resources)

        self.html.append("<h3>Result</h3>")
        if item.result_is_file:
            self.html.append("<i>File</i><br>")
            self.add_file(item.result)
        elif item.result_is_array_of_files:
            self.html.append("<i>Array of files</i><br>")
            for file in item.result:
                self.add_file(file)
        elif isinstance(item.result, list) or isinstance(item.result, dict):
            self.add_code(json.dumps(item.result, indent=2))
        else:
            self.add_code(str(item.result))

    def process_source_code(self, code: str):
        pattern = r"\.href\((['\"])(.*?)\1\)"
        self.html.append('<p style="font-family: monospace;">')
        for line in code.split('\n'):
            line = self.replace_code(line)
            self.html.append(line)
            match = re.search(pattern, line)
            if match is not None:
                self.html.append(f'<a href="#{match.group(2)}">[→]</a>')
            self.html.append("<br>")


    def build(self, report: TestReport):
        self.html.append(f'<h1>Test source code</h1>')
        for index, code in enumerate(report.source_codes):
            if code is None:
                raise ValueError(f"Code is none at index {index}")
            self.process_source_code(code)


        for section in report.sections:
            self.html.append(f'<h1>{section.caption}</h1>')
            if section.comment is not None:
                self.html.append(f'<p>{section.comment}</p>')
            for item in section.items:
                self.process_item(item)

        if report.error is not None:
            self.html.append(f"<h1 {self.error_color_style()}><b>Exception</b></h1>")
            self.add_code(report.error, self.error_color_style())

        for i in range(len(self.html)):
            if self.html[i] is None:
                print(self.html[max(0,i-5):min(len(self.html), i+5)])
                raise ValueError(f"None in html at index {i}")
        return "<html><body>"+"".join(self.html)+"</body></html>"





def create_self_test_report_page(report: TestReport):
    builder = Builder()
    return builder.build(report)


from typing import *
from uuid import uuid4
from copy import copy
from ...common import SignatureProcessor
from ..app import IBrainBoxTask
from ...job_processing import Job
from .task_builder import BrainBoxTaskBuilderResult, BrainBoxTaskBuilder
from .helper_methods import fix_decider_and_decider_method, fix_arguments_and_dependencies

TDecider = TypeVar('TDecider')
TValue = TypeVar('TValue')




class BrainBoxTask(IBrainBoxTask):
    def __init__(self,
                 *,
                 decider: Union[str,type,Callable],
                 arguments: Optional[Dict[str,Any]] = None,
                 id: str | None = None,
                 dependencies: Optional[Dict[str,Union[str, 'BrainBoxTask']]] = None,
                 fake_dependencies: Optional[List[Union[str,'BrainBoxTask']]] = None,
                 info: Any = None,
                 batch: str|None = None,
                 decider_method: Union[None, str, Callable] = None,
                 decider_parameter: None|str = None,
                 ordering_token: str|None = None,
                 ):
        if id is not None:
            self.id = id
        else:
            self.id = BrainBoxTask.safe_id()

        self.decider, self.decider_method, _method_to_validate = fix_decider_and_decider_method(decider, decider_method)

        self.arguments, self.dependencies = fix_arguments_and_dependencies(arguments, dependencies, fake_dependencies)
        if _method_to_validate is not None:
            SignatureProcessor.from_signature(_method_to_validate).to_kwargs_only(**self.arguments, **self.dependencies)

        self.info = info
        self.batch = batch
        self.decider_parameter = decider_parameter
        self.ordering_token = ordering_token
        


    def __repr__(self):
        return self.__dict__.__repr__()

    @staticmethod
    def call(_type: Type[TDecider], parameter: str|None = None) -> TDecider:
        return BrainBoxTaskBuilder(_type, parameter=parameter)

    @staticmethod
    def build(call_result: Any) -> BrainBoxTaskBuilderResult:
        return call_result


    def clone(self, **kwargs):
        result = copy(self)
        for k, v in kwargs.items():
            setattr(result, k, v)
        return result

    def get_resulting_id(self) -> str:
        return self.id

    def create_jobs(self) -> list[Job]:
        job = Job(
            id  = self.id,
            decider = self.decider,
            decider_parameter = self.decider_parameter,
            method = self.decider_method,
            arguments = self.arguments,
            info = self.info,
            batch = self.batch,
            ordering_token = self.ordering_token,
            dependencies = self.dependencies
        )
        return [job]





from typing import *
from ..planner import IPlanner, PlannerArguments, StartCommand, StopCommand, AssignAction
from ..core import DeciderInstanceKey, OperatorStateForPlanner
from yo_fluq import *
from datetime import datetime

class SimplePlanner(IPlanner):
    def __init__(self,
                 cooldown_delay_in_seconds: int|None = 60*30,
                 datetime_factory: Optional[Callable[[], datetime]] = datetime.now
                 ):
        self.cooldown_delay_in_seconds = cooldown_delay_in_seconds
        self.datetime_factory = datetime_factory
        self.has_cooldown_delay = self.cooldown_delay_in_seconds is not None and self.datetime_factory is not None

    def _cooldown_respecting_delay(self, active_service: OperatorStateForPlanner):
        if not self.has_cooldown_delay:
            return [StopCommand(active_service.instance_id)]
        now = self.datetime_factory()
        if active_service.not_busy_since is None:
            return [StopCommand(active_service.instance_id)]
        delta = (now - active_service.not_busy_since).total_seconds()
        if delta < self.cooldown_delay_in_seconds:
            return []
        return [StopCommand(active_service.instance_id)]

    def _choose_next_decider_to_activate(self, args: PlannerArguments):
        service_to_activate: list[DeciderInstanceKey] = (Query
                    .en(args.non_finished_tasks)
                    .group_by(lambda z: z.get_decider_instance_key())
                    .select(lambda z: (z.key, len(z.value)))
                    .order_by_descending(lambda z: z[1])
                    .then_by(lambda z: str(z[0]))
                    .select(lambda z: z[0])
                    .to_list()
                    )
        if len(service_to_activate) == 0:
            return []
        else:
            args.log_handler.event(f"No active services, starting {service_to_activate[0]}")
            return [StartCommand(service_to_activate[0], StartCommand.Mode.StartOnly)]

    def _get_tasks_for_service(self, args: PlannerArguments, active_service: OperatorStateForPlanner):
        return (
            Query
            .en(args.non_finished_tasks)
            .where(lambda z: z.get_decider_instance_key() == active_service.key)
            .to_list()
        )



    def plan(self, args: PlannerArguments):
        active_service: OperatorStateForPlanner|None = None

        if len(args.deciders) > 1:
            raise ValueError("Simple planner cannot operate with more than 1 decider active")
        elif len(args.deciders) == 1:
            active_service = args.deciders[0]

        if active_service is None:
            return self._choose_next_decider_to_activate(args)

        tasks_for_active_service = self._get_tasks_for_service(args, active_service)
        currently_at_active_service = Query.en(tasks_for_active_service).where(lambda z: z.assigned).to_list()
        next_for_active_service = (
            Query
            .en(tasks_for_active_service)
            .where(lambda z: not z.assigned)
            .order_by(lambda z: z.ordering_token)
            .then_by(lambda z: z.received_timestamp)
            .to_list()
        )

        if len(tasks_for_active_service) == 0:
            if len(args.non_finished_tasks) > 0:
                args.log_handler.event("No tasks for active service, there are tasks for other services, stopping")
                return [StopCommand(active_service.instance_id)]
            result = self._cooldown_respecting_delay(active_service)
            if len(result) > 0:
                args.log_handler.event("No tasks, cooldown time waited, stopping")
            return result

        if len(currently_at_active_service) >= 2:
            return []

        if len(next_for_active_service) == 0:
            return []

        next = next_for_active_service[0]
        args.log_handler.event(f'Assigning task {next.id}')
        return [AssignAction(next.id, active_service.instance_id, active_service.key)]











from typing import Callable
from sqlalchemy.orm import Session
from sqlalchemy import event
from dataclasses import dataclass, field

class TrackableSessionFactory:
    @dataclass
    class Changes:
        modified: list = field(default_factory=list)
        added: list = field(default_factory=list)
        deleted: list = field(default_factory=list)

        @property
        def all(self) -> list:
            return self.modified+self.added+self.deleted

    def __init__(self, engine, selector:Callable):
        self.engine = engine
        self.selector = selector
        self.changes = TrackableSessionFactory.Changes()

    def __call__(self) -> Session:
        session = Session(self.engine)
        event.listen(session, 'before_commit', self._track_changes)
        return session

    def _track_changes(self, session):
        self.changes.modified.extend(self.selector(obj) for obj in session.dirty if session.is_modified(obj, include_collections=False))
        self.changes.added.extend(self.selector(obj) for obj in session.new)
        self.changes.deleted.extend(self.selector(obj) for obj in session.deleted)

    def reset(self):
        self.changes = TrackableSessionFactory.Changes()


from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import WD14TaggerModel

@dataclass
class WD14TaggerSettings:
    connection = ConnectionSettings(20303)
    models_to_download: tuple[WD14TaggerModel,...] = (
        WD14TaggerModel('wd14-vit.v2', 'models--SmilingWolf--wd-v1-4-vit-tagger-v2'),
    )
    cpu_share:float = 0.25



from typing import *
from sqlalchemy.orm import declarative_base, Mapped, mapped_column
from sqlalchemy import JSON, PickleType, func
from datetime import datetime
from .decider_instance_key import DeciderInstanceKey

BrainBoxBase = declarative_base()

class Job(BrainBoxBase):
    __tablename__ = "brain_box_jobs"

    id: Mapped[str] = mapped_column(primary_key=True)
    decider: Mapped[str] = mapped_column()
    decider_parameter: Mapped[str] = mapped_column(nullable=True)
    method: Mapped[str] = mapped_column(nullable=True)
    arguments: Mapped[Any] = mapped_column(type_ = PickleType)
    info: Mapped[Any] = mapped_column(type_= PickleType, nullable=True)
    batch: Mapped[str] = mapped_column()
    ordering_token: Mapped[str] = mapped_column(nullable=True)

    received_timestamp: Mapped[datetime] = mapped_column()

    dependencies: Mapped[Optional[Dict[str, str]]] = mapped_column(type_=JSON, default = None)
    has_dependencies: Mapped[bool] = mapped_column(nullable=False)
    ready: Mapped[bool] = mapped_column(default=False)
    ready_timestamp: Mapped[datetime] = mapped_column(nullable=True)

    assigned: Mapped[bool] = mapped_column(default=False)
    assigned_timestamp: Mapped[datetime] = mapped_column(nullable=True)

    accepted: Mapped[bool] = mapped_column(default=False)
    accepted_timestamp: Mapped[datetime] = mapped_column(nullable=True)

    progress: Mapped[Optional[float]] = mapped_column(nullable=True, default=None)
    log: Mapped[Optional[List[str]]] = mapped_column(nullable=True, default=None, type_ = JSON)

    finished: Mapped[bool] = mapped_column(default=False)
    finished_timestamp: Mapped[datetime] = mapped_column(nullable=True)
    success: Mapped[bool] = mapped_column(default=False)
    result: Mapped[Any] = mapped_column(type_ = PickleType, nullable=True, default=None)
    error: Mapped[str] = mapped_column(nullable=True, default=None)

    last_update_timestamp: Mapped[datetime] = mapped_column(
        default=func.strftime('%Y-%m-%d %H:%M:%f', 'now'),
        onupdate=func.strftime('%Y-%m-%d %H:%M:%f', 'now')
    )


    def get_this_and_dependency_ids(self):
        result = [self.id]
        if self.dependencies is not None:
            for key in sorted(self.dependencies):
                result.append(self.dependencies[key])
        return result

    def get_key(self):
        return DeciderInstanceKey(self.decider, self.decider_parameter)

    def set_defaults(self, now: datetime|None = None):
        if now is None:
            now = datetime.now()
        self.received_timestamp = now
        self.batch = self.batch if self.batch is not None else self.id
        self.has_dependencies = self.dependencies is not None
        return self




from typing import *
from dataclasses import dataclass
from pathlib import Path

@dataclass
class Command:
    @dataclass
    class Options:
        workdir: str | Path | None = None
        return_output: bool = False
        ignore_exit_code: bool = False
        monitor_output: Optional[Callable[[str], None]] = None


    command: tuple[str,...]
    options: Optional['Command.Options'] = None




import os
import shutil

import requests
from ....framework import DockerWebServiceApi, File, FileLike
from .controller import OpenVoiceController
from .settings import OpenVoiceSettings
from typing import Iterable


class OpenVoice(DockerWebServiceApi[OpenVoiceSettings, OpenVoiceController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def generate(self, source: FileLike.Type, voice: str, source_voice: str|None = None) -> File:
        url = f"http://{self.address}/generate/{voice}"
        if source_voice is not None:
            url+='/'+source_voice
        with FileLike(source, self.cache_folder) as stream:
            reply = requests.post(url, files=(('file', stream),), timeout=30)
        if reply.status_code != 200:
            raise ValueError(f"{reply.status_code}\n{reply.text}")
        
        return File(self.current_job_id+'.output.wav', reply.content, File.Kind.Audio)

    def train(self, voice: str, files: Iterable[FileLike.Type]):
        folder = self.controller.resource_folder('voices',voice)
        shutil.rmtree(folder, ignore_errors=True)
        os.makedirs(folder)
        for file in files:
            with open(folder/FileLike.get_name(file), 'wb') as stream:
                with FileLike(file, self.cache_folder) as data:
                    stream.write(data.read())
        reply = requests.post(self.endpoint(f'train/{voice}'), json=dict(voice=voice))
        if reply.status_code!=200:
            raise ValueError(f'Endpoint train raised for voice {voice}\n{reply.text}')
        return "OK"


    Settings = OpenVoiceSettings
    Controller = OpenVoiceController

    



from pathlib import Path

from ....framework import DownloadableModel, IController
from dataclasses import dataclass
from typing import cast

@dataclass
class YoloModel(DownloadableModel):
    repo_id: str
    file_name: str

    @property
    def model_id(self):
        return f'{self.repo_id}:{self.file_name}'

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('models')/self.repo_id/self.file_name

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import YoloController
        from .api import Yolo
        controller = cast(YoloController, controller)
        instance_id = controller.run(None)
        api: Yolo = controller.find_api(instance_id)
        for model in models:
            model = cast(YoloModel, model)
            api.load_model(model.model_id)
        controller.stop(instance_id)



import os
from ...framework import IDecider, File
import json
from copy import copy

class CacheManager(IDecider):
    def delete(self, files: str|list[str]):
        if isinstance(files,str):
            files = [files]
        for file in files:
            os.unlink(self.cache_folder/file)



from .marshalling_metadata import MarshallingMetadata
from .format import Format
import requests

class ApiBinding:
    def __init__(self,
                 address: str,
                 metadata: MarshallingMetadata,
                 ):
        self.metadata = metadata
        self.address = address

    def __call__(self, *args, **kwargs):
        address = f'http://{self.address}{self.metadata.get_endpoint_address()}'
        arguments = self.metadata.signature.to_kwargs_only(*args, **kwargs)
        data = dict(
            arguments = Format.encode(arguments),
        )

        reply = requests.request(
            self.metadata.endpoint.method,
            address,
            json = data,
        )
        if reply.status_code == 200:
            result = Format.decode(reply.json())
            return result['result']
        try:
            error = reply.json()
        except:
            raise ValueError(f"Call to {address} caused unprocessed error on the server\n\n{reply.text}")
        if 'error' not in error:
            raise ValueError(f"Call to {address} caused unprocessed error on the server\n\n{error}")
        raise ValueError(f"Call to {address} caused exception:\n\n{error['error']}")




import os
import subprocess
import logging
import traceback

import flask
from flask import Flask, request, jsonify, send_from_directory
from werkzeug.utils import secure_filename
import torch
from resemble_enhance.enhancer.inference import denoise, enhance
import torchaudio


UPLOAD_FOLDER = '/uploads'
OUTPUT_FOLDER = '/outputs'

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

class DenoiserApp:

    def create_app(self):
        self.app = Flask(__name__)
        self.app.add_url_rule('/', view_func=self.index, methods=['GET'])
        self.app.add_url_rule('/enhance', view_func=self.enhance_audio, methods=['POST'])
        return self.app

    def index(self):
        return "OK"

    def enhance_audio(self):
        try:
            filename = request.json['filename']
            solver = request.json.get('solver', 'Midpoint') #"CFM ODE Solver", Options ["Midpoint", "RK4", "Euler"]
            nfe = request.json.get('nfe', 64) #"CFM Number of Function Evaluations", from 1 to 128
            tau = request.json.get('tau', 0.5) #"CFM Prior Temperature", from 0 to 1
            denoising = request.json.get('denoising', False) #Denoise before enhancement

            path = os.path.join(UPLOAD_FOLDER, filename)

            solver = solver.lower()
            nfe = int(nfe)
            lambd = 0.9 if denoising else 0.1

            dwav, sr = torchaudio.load(path)
            dwav = dwav.mean(dim=0)

            wav1, new_sr = denoise(dwav, sr, device)
            wav2, new_sr = enhance(dwav, sr, device, nfe=nfe, solver=solver, lambd=lambd, tau=tau)

            wav1 = wav1.cpu().numpy()
            wav2 = wav2.cpu().numpy()

            wav1 = torch.from_numpy(wav1).unsqueeze(0)
            wav2 = torch.from_numpy(wav2).unsqueeze(0)

            denoised_name = filename+'.denoised.wav'
            torchaudio.save(os.path.join(OUTPUT_FOLDER, denoised_name), wav1, new_sr)
            enhanced_name = filename+'.enhanced.wav'
            torchaudio.save(os.path.join(OUTPUT_FOLDER, enhanced_name), wav2, new_sr)
            return flask.jsonify([denoised_name, enhanced_name])
        except:
            return traceback.format_exc(), 500


